/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=True $ backoff_blocking_factor=4 $ parcomb=True $ np_schedulers=set(['fg']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=8 $ instrument=set([]) $ unroll=['pipes'] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=True $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
typedef int edge_data_type;
typedef int node_data_type;
#define RESIDENCY 6
extern const node_data_type INF = INT_MAX;
static const int __tb_gg_main_pipe_0_gpu_gb = 256;
static const int __tb_contract_expand = TB_SIZE;
static const int __tb_one = 1;
__device__ void kernel_dev(CSRGraph graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = INF ;
  }
  // FP: "4 -> 5;
}
__global__ void kernel(CSRGraph graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  // FP: "1 -> 2;
  kernel_dev(graph, src);
  // FP: "2 -> 3;
}
__device__ void contract_expand_dev(CSRGraph graph, int LEVEL, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_contract_expand;
  index_type wlnode_end;
  index_type wlnode_rup;
  // FP: "1 -> 2;
  const int _NP_CROSSOVER_WP = 32;
  const int _NP_CROSSOVER_TB = __kernel_tb_size;
  // FP: "2 -> 3;
  const int BLKSIZE = __kernel_tb_size;
  const int ITSIZE = BLKSIZE * 8;
  // FP: "3 -> 4;

  typedef cub::BlockScan<index_type, BLKSIZE> BlockScan;
  typedef union np_shared<BlockScan::TempStorage, index_type, struct empty_np, struct empty_np, struct fg_np<ITSIZE> > npsTy;

  // FP: "4 -> 5;
  __shared__ npsTy nps ;
  // FP: "5 -> 6;
  const int WHTSIZE = 64;
  __shared__ volatile int whash[__kernel_tb_size/32][WHTSIZE] ;
  // FP: "6 -> 7;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  wlnode_rup = ((0) + roundup(((*((volatile index_type *) (in_wl).dindex)) - (0)), (blockDim.x)));
  for (index_type wlnode = 0 + tid; wlnode < wlnode_rup; wlnode += nthreads)
  {
    int node;
    bool pop;
    index_type _start_43;
    index_type _np_mps;
    index_type _np_mps_total;
    __shared__ index_type _np_pc__start_43;
    // FP: "7 -> 8;
    // FP: "8 -> 9;
    pop = (in_wl).pop_id(wlnode, node);
    // FP: "9 -> 10;
    if (pop)
    {
      pop = false;
      if (graph.node_data[node] == INF)
      {
        graph.node_data[node] = LEVEL;
        pop = true;
      }
    }
    // FP: "15 -> 16;
    if (pop)
    {
      const int warpid = threadIdx.x / 32;
      const int htentry = node & (WHTSIZE - 1);
      whash[warpid][htentry] = node;
      if (whash[warpid][htentry] == node)
      {
        whash[warpid][htentry] = threadIdx.x;
        pop = whash[warpid][htentry] == threadIdx.x;
      }
    }
    // FP: "22 -> 23;
    __syncthreads();
    // FP: "23 -> 24;
    // FP: "26 -> 27;
    // FP: "27 -> 28;
    struct NPInspector1 _np = {0,0,0,0,0,0};
    // FP: "28 -> 29;
    if (pop)
    {
      _np.size = (graph).getOutDegree(node);
      _np.start = (graph).getFirstEdge(node);
    }
    // FP: "31 -> 32;
    // FP: "32 -> 33;
    _np_mps = _np.size;
    // FP: "33 -> 34;
    BlockScan(nps.temp_storage).ExclusiveSum(_np_mps, _np_mps, _np_mps_total);
    // FP: "34 -> 35;
    // FP: "35 -> 36;
    if (threadIdx.x == 0)
    {
      _np_pc__start_43 = (out_wl).push_range(_np_mps_total);;
    }
    // FP: "38 -> 39;
    __syncthreads();
    // FP: "39 -> 40;
    _np.total = _np_mps_total;
    _np.offset = _np_mps;
    // FP: "40 -> 41;
    _start_43 = _np_pc__start_43;
    // FP: "41 -> 42;
    while (_np.work())
    {
      // FP: "42 -> 43;
      int _np_i =0;
      // FP: "43 -> 44;
      _np.inspect(nps.fg.itvalue, ITSIZE);
      // FP: "44 -> 45;
      __syncthreads();
      // FP: "45 -> 46;

      // FP: "46 -> 47;
      for (_np_i = threadIdx.x; _np_i < ITSIZE && _np.valid(_np_i); _np_i += BLKSIZE)
      {
        index_type edge;
        index_type edge_pos  = _np_i;
        edge= nps.fg.itvalue[_np_i];
        {
          index_type dst;
          dst = graph.getAbsDestination(edge);
          (out_wl).do_push(_start_43, edge_pos, dst);
        }
      }
      // FP: "54 -> 55;
      _np.execute_round_done(ITSIZE);
      // FP: "55 -> 56;
      _start_43 = _start_43 + ITSIZE;;
      // FP: "56 -> 57;
      __syncthreads();
    }
  }
  // FP: "59 -> 60;
}
__global__ void __launch_bounds__(TB_SIZE, RESIDENCY) contract_expand(CSRGraph graph, int LEVEL, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_contract_expand;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  contract_expand_dev(graph, LEVEL, in_wl, out_wl);
  // FP: "2 -> 3;
}
void gg_main_pipe_0(CSRGraph& gg, int& LEVEL, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  // FP: "1 -> 2;
  {
    kernel <<<blocks, threads>>>(gg, 0);
    // FP: "3 -> 4;
    while (pipe.in_wl().nitems())
    {
      pipe.out_wl().will_write();
      contract_expand <<<blocks, __tb_contract_expand>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      LEVEL++;
    }
    // FP: "7 -> 8;
  }
  // FP: "2 -> 3;
}
__global__ void __launch_bounds__(__tb_gg_main_pipe_0_gpu_gb) gg_main_pipe_0_gpu_gb(CSRGraph gg, int LEVEL, PipeContextT<Worklist2> pipe, int* cl_LEVEL, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_gg_main_pipe_0_gpu_gb;
  // FP: "1 -> 2;
  LEVEL = *cl_LEVEL;
  // FP: "2 -> 3;
  {
    kernel_dev (gg, 0);
    gb.Sync();
    // FP: "4 -> 5;
    while (pipe.in_wl().nitems())
    {
      if (tid == 0)
        pipe.in_wl().reset_next_slot();
      contract_expand_dev (gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      gb.Sync();
      pipe.advance2();
      LEVEL++;
      if (tid == 0)
        pipe.in_wl().reset_next_slot();
      contract_expand_dev (gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      gb.Sync();
      pipe.advance2();
      LEVEL++;
      if (tid == 0)
        pipe.in_wl().reset_next_slot();
      contract_expand_dev (gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      gb.Sync();
      pipe.advance2();
      LEVEL++;
      if (tid == 0)
        pipe.in_wl().reset_next_slot();
      contract_expand_dev (gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      gb.Sync();
      pipe.advance2();
      LEVEL++;
    }
    // FP: "14 -> 15;
  }
  // FP: "3 -> 4;
  gb.Sync();
  // FP: "15 -> 16;
  if (tid == 0)
  {
    *cl_LEVEL = LEVEL;
  }
  // FP: "18 -> 19;
}
__global__ void gg_main_pipe_0_gpu(CSRGraph gg, int LEVEL, PipeContextT<Worklist2> pipe, dim3 blocks, dim3 threads, int* cl_LEVEL)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_one;
  // FP: "1 -> 2;
  LEVEL = *cl_LEVEL;
  // FP: "2 -> 3;
  {
    kernel <<<blocks, threads>>>(gg, 0);
    cudaDeviceSynchronize();
    // FP: "4 -> 5;
    while (pipe.in_wl().nitems())
    {
      contract_expand <<<blocks, __tb_contract_expand>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      cudaDeviceSynchronize();
      pipe.advance2();
      LEVEL++;
      contract_expand <<<blocks, __tb_contract_expand>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      cudaDeviceSynchronize();
      pipe.advance2();
      LEVEL++;
      contract_expand <<<blocks, __tb_contract_expand>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      cudaDeviceSynchronize();
      pipe.advance2();
      LEVEL++;
      contract_expand <<<blocks, __tb_contract_expand>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      cudaDeviceSynchronize();
      pipe.advance2();
      LEVEL++;
    }
    // FP: "14 -> 15;
  }
  // FP: "3 -> 4;
  if (tid == 0)
  {
    *cl_LEVEL = LEVEL;
  }
  // FP: "17 -> 18;
}
void gg_main_pipe_0_wrapper(CSRGraph& gg, int& LEVEL, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  static GlobalBarrierLifetime gg_main_pipe_0_gpu_gb_barrier;
  static bool gg_main_pipe_0_gpu_gb_barrier_inited;
  // FP: "1 -> 2;
  static const size_t gg_main_pipe_0_gpu_gb_residency = maximum_residency(gg_main_pipe_0_gpu_gb, __tb_gg_main_pipe_0_gpu_gb, 0);
  static const size_t gg_main_pipe_0_gpu_gb_blocks = GG_MIN(blocks.x, ggc_get_nSM() * gg_main_pipe_0_gpu_gb_residency);
  if(!gg_main_pipe_0_gpu_gb_barrier_inited) { gg_main_pipe_0_gpu_gb_barrier.Setup(gg_main_pipe_0_gpu_gb_blocks); gg_main_pipe_0_gpu_gb_barrier_inited = true;};
  // FP: "2 -> 3;
  if (false)
  {
    gg_main_pipe_0(gg,LEVEL,pipe,blocks,threads);
  }
  else
  {
    int* cl_LEVEL;
    check_cuda(cudaMalloc(&cl_LEVEL, sizeof(int) * 1));
    check_cuda(cudaMemcpy(cl_LEVEL, &LEVEL, sizeof(int) * 1, cudaMemcpyHostToDevice));

    // gg_main_pipe_0_gpu<<<1,1>>>(gg,LEVEL,pipe,blocks,threads,cl_LEVEL);
    gg_main_pipe_0_gpu_gb<<<gg_main_pipe_0_gpu_gb_blocks, __tb_gg_main_pipe_0_gpu_gb>>>(gg,LEVEL,pipe,cl_LEVEL, gg_main_pipe_0_gpu_gb_barrier);
    check_cuda(cudaMemcpy(&LEVEL, cl_LEVEL, sizeof(int) * 1, cudaMemcpyDeviceToHost));
    check_cuda(cudaFree(cl_LEVEL));
  }
  // FP: "5 -> 6;
}
void gg_main(CSRGraph& hg, CSRGraph& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  static GlobalBarrierLifetime contract_expand_barrier;
  static bool contract_expand_barrier_inited;
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  static const size_t contract_expand_residency = maximum_residency(contract_expand, __tb_contract_expand, 0);
  static const size_t contract_expand_blocks = GG_MIN(blocks.x, ggc_get_nSM() * contract_expand_residency);
  if(!contract_expand_barrier_inited) { contract_expand_barrier.Setup(contract_expand_blocks); contract_expand_barrier_inited = true;};
  // FP: "2 -> 3;
  int LEVEL = 0;
  // FP: "3 -> 4;
  pipe = PipeContextT<Worklist2>(gg.nedges > 65536 ? gg.nedges : 65536);
  pipe.in_wl().wl[0] = 0;
  pipe.in_wl().update_gpu(1);
  gg_main_pipe_0_wrapper(gg,LEVEL,pipe,blocks,threads);
  pipe.free();
  // FP: "4 -> 5;
}