/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=True $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=8 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=True $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
typedef int edge_data_type;
typedef int node_data_type;
extern const node_data_type INF = INT_MAX;
static const int __tb_filter = TB_SIZE;
static const int __tb_contract = TB_SIZE;
static const int __tb_expand = TB_SIZE;
__global__ void kernel(CSRGraph graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = INF ;
  }
  // FP: "4 -> 5;
}
__global__ void __launch_bounds__(TB_SIZE, 8) expand(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_expand;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  index_type wlnode_rup;
  // FP: "1 -> 2;
  const int _NP_CROSSOVER_WP = 32;
  const int _NP_CROSSOVER_TB = __kernel_tb_size;
  // FP: "2 -> 3;
  const int BLKSIZE = __kernel_tb_size;
  const int ITSIZE = BLKSIZE * 8;
  // FP: "3 -> 4;

  typedef cub::BlockScan<multiple_sum<2, index_type>, BLKSIZE> BlockScan;
  typedef union np_shared<BlockScan::TempStorage, index_type, struct tb_np, struct warp_np<__kernel_tb_size/32>, struct fg_np<ITSIZE> > npsTy;

  // FP: "4 -> 5;
  __shared__ npsTy nps ;
  // FP: "5 -> 6;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  wlnode_rup = ((0) + roundup(((*((volatile index_type *) (in_wl).dindex)) - (0)), (blockDim.x)));
  for (index_type wlnode = 0 + tid; wlnode < wlnode_rup; wlnode += nthreads)
  {
    int node;
    bool pop;
    index_type _start_20;
    multiple_sum<2, index_type> _np_mps;
    multiple_sum<2, index_type> _np_mps_total;
    __shared__ index_type _np_pc__start_20;
    // FP: "6 -> 7;
    // FP: "7 -> 8;
    pop = (in_wl).pop_id(wlnode, node);
    // FP: "8 -> 9;
    // FP: "9 -> 10;
    struct NPInspector1 _np = {0,0,0,0,0,0};
    // FP: "10 -> 11;
    if (pop)
    {
      _np.size = (graph).getOutDegree(node);
      _np.start = (graph).getFirstEdge(node);
    }
    // FP: "13 -> 14;
    // FP: "14 -> 15;
    _np_mps.el[0] = _np.size >= _NP_CROSSOVER_WP ? _np.size : 0;
    _np_mps.el[1] = _np.size < _NP_CROSSOVER_WP ? _np.size : 0;
    // FP: "15 -> 16;
    BlockScan(nps.temp_storage).ExclusiveSum(_np_mps, _np_mps, _np_mps_total);
    // FP: "16 -> 17;
    // FP: "17 -> 18;
    if (threadIdx.x == 0)
    {
      nps.tb.owner = MAX_TB_SIZE + 1;
      _np_pc__start_20 = (out_wl).push_range(_np_mps_total.el[0] + _np_mps_total.el[1]);;
    }
    // FP: "21 -> 22;
    __syncthreads();
    // FP: "22 -> 23;
    while (true)
    {
      // FP: "23 -> 24;
      if (_np.size >= _NP_CROSSOVER_TB)
      {
        nps.tb.owner = threadIdx.x;
      }
      // FP: "26 -> 27;
      __syncthreads();
      // FP: "27 -> 28;
      if (nps.tb.owner == MAX_TB_SIZE + 1)
      {
        // FP: "28 -> 29;
        __syncthreads();
        // FP: "29 -> 30;
        break;
      }
      // FP: "31 -> 32;
      if (nps.tb.owner == threadIdx.x)
      {
        nps.tb.offset = _np_mps.el[0];
        nps.tb.start = _np.start;
        nps.tb.size = _np.size;
        nps.tb.src = threadIdx.x;
        _np.start = 0;
        _np.size = 0;
      }
      // FP: "35 -> 36;
      __syncthreads();
      // FP: "36 -> 37;
      int ns = nps.tb.start;
      int ne = nps.tb.size;
      // FP: "37 -> 38;
      if (nps.tb.src == threadIdx.x)
      {
        nps.tb.owner = MAX_TB_SIZE + 1;
      }
      // FP: "40 -> 41;
      _start_20 = _np_pc__start_20 + nps.tb.offset;
      // FP: "41 -> 42;
      for (int _np_j = threadIdx.x; _np_j < ne; _np_j += BLKSIZE)
      {
        index_type edge;
        edge = ns +_np_j;
        index_type edge_pos = _np_j;
        {
          index_type dst;
          dst = graph.getAbsDestination(edge);
          (out_wl).do_push(_start_20, edge_pos, dst);
        }
      }
      // FP: "49 -> 50;
      __syncthreads();
    }
    // FP: "51 -> 52;

    // FP: "52 -> 53;
    {
      const int warpid = threadIdx.x / 32;
      // FP: "53 -> 54;
      const int _np_laneid = cub::LaneId();
      // FP: "54 -> 55;
      while (__any(_np.size >= _NP_CROSSOVER_WP && _np.size < _NP_CROSSOVER_TB))
      {
        if (_np.size >= _NP_CROSSOVER_WP && _np.size < _NP_CROSSOVER_TB)
        {
          nps.warp.owner[warpid] = _np_laneid;
        }
        if (nps.warp.owner[warpid] == _np_laneid)
        {
          nps.warp.offset[warpid] = _np_mps.el[0];
          nps.warp.start[warpid] = _np.start;
          nps.warp.size[warpid] = _np.size;

          _np.start = 0;
          _np.size = 0;
        }
        index_type _np_w_start = nps.warp.start[warpid];
        index_type _np_w_size = nps.warp.size[warpid];
        _start_20 =  _np_pc__start_20 + nps.warp.offset[warpid];
        for (int _np_ii = _np_laneid; _np_ii < _np_w_size; _np_ii += 32)
        {
          index_type edge;
          edge = _np_w_start +_np_ii;
          index_type edge_pos = _np_ii;
          {
            index_type dst;
            dst = graph.getAbsDestination(edge);
            (out_wl).do_push(_start_20, edge_pos, dst);
          }
        }
      }
      // FP: "73 -> 74;
      __syncthreads();
      // FP: "74 -> 75;
    }

    // FP: "75 -> 76;
    __syncthreads();
    // FP: "76 -> 77;
    _np.total = _np_mps_total.el[1];
    _np.offset = _np_mps.el[1];
    // FP: "77 -> 78;
    _start_20 = _np_pc__start_20 + _np_mps_total.el[0];
    // FP: "78 -> 79;
    while (_np.work())
    {
      // FP: "79 -> 80;
      int _np_i =0;
      // FP: "80 -> 81;
      _np.inspect(nps.fg.itvalue, ITSIZE);
      // FP: "81 -> 82;
      __syncthreads();
      // FP: "82 -> 83;

      // FP: "83 -> 84;
      for (_np_i = threadIdx.x; _np_i < ITSIZE && _np.valid(_np_i); _np_i += BLKSIZE)
      {
        index_type edge;
        index_type edge_pos  = _np_i;
        edge= nps.fg.itvalue[_np_i];
        {
          index_type dst;
          dst = graph.getAbsDestination(edge);
          (out_wl).do_push(_start_20, edge_pos, dst);
        }
      }
      // FP: "91 -> 92;
      _np.execute_round_done(ITSIZE);
      // FP: "92 -> 93;
      _start_20 = _start_20 + ITSIZE;;
      // FP: "93 -> 94;
      __syncthreads();
    }
  }
  // FP: "96 -> 97;
}
__global__ void __launch_bounds__(TB_SIZE, 8) contract(CSRGraph graph, int LEVEL, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop)
    {
      if (graph.node_data[node] == INF)
      {
        index_type _start_32;
        graph.node_data[node] = LEVEL;
        _start_32 = (out_wl).setup_push_warp_one();;
        (out_wl).do_push(_start_32, 0, node);
      }
    }
  }
  // FP: "13 -> 14;
}
__global__ void __launch_bounds__(TB_SIZE, 8) filter(ApproxBitsetByte visited, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop)
    {
      if (!visited.is_set(node))
      {
        index_type _start_47;
        visited.set(node);
        _start_47 = (out_wl).setup_push_warp_one();;
        (out_wl).do_push(_start_47, 0, node);
      }
    }
  }
  // FP: "13 -> 14;
}
void gg_main(CSRGraph& hg, CSRGraph& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  static GlobalBarrierLifetime expand_barrier;
  static bool expand_barrier_inited;
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  static const size_t expand_residency = maximum_residency(expand, __tb_expand, 0);
  static const size_t expand_blocks = GG_MIN(blocks.x, ggc_get_nSM() * expand_residency);
  if(!expand_barrier_inited) { expand_barrier.Setup(expand_blocks); expand_barrier_inited = true;};
  // FP: "2 -> 3;
  kernel <<<blocks, threads>>>(gg, 0);
  // FP: "3 -> 4;
  int LEVEL = 0;
  // FP: "4 -> 5;
  ApproxBitsetByte visited (hg.nnodes);
  // FP: "5 -> 6;
  pipe = PipeContextT<Worklist2>(gg.nedges > 65536 ? gg.nedges : 65536);
  pipe.in_wl().wl[0] = 0;
  pipe.in_wl().update_gpu(1);
  {
    while (pipe.in_wl().nitems())
    {
      pipe.out_wl().will_write();
      contract <<<blocks, __tb_contract>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      pipe.out_wl().will_write();
      expand <<<blocks, __tb_expand>>>(gg, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      pipe.out_wl().will_write();
      filter <<<blocks, __tb_filter>>>(visited, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      LEVEL++;
    }
  }
  pipe.free();
  // FP: "11 -> 12;
}