/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=True $ backoff_blocking_factor=4 $ parcomb=True $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=8 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=True $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
typedef int edge_data_type;
typedef int node_data_type;
extern const node_data_type INF = INT_MAX;
static const int __tb_bfs_kernel = TB_SIZE;
static const int __tb_one = 1;
static const int __tb_gg_main_pipe_1_gpu_gb = 256;
__global__ void bfs_init(CSRGraph graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = (node == src) ? 0 : INF ;
  }
  // FP: "4 -> 5;
}
__device__ void bfs_kernel_dev(CSRGraph graph, int LEVEL, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_bfs_kernel;
  index_type wlnode_end;
  index_type wlnode_rup;
  // FP: "1 -> 2;
  const int _NP_CROSSOVER_WP = 32;
  const int _NP_CROSSOVER_TB = __kernel_tb_size;
  // FP: "2 -> 3;
  const int BLKSIZE = __kernel_tb_size;
  const int ITSIZE = BLKSIZE * 8;
  // FP: "3 -> 4;

  typedef cub::BlockScan<multiple_sum<2, index_type>, BLKSIZE> BlockScan;
  typedef union np_shared<BlockScan::TempStorage, index_type, struct tb_np, struct warp_np<__kernel_tb_size/32>, struct fg_np<ITSIZE> > npsTy;

  // FP: "4 -> 5;
  __shared__ npsTy nps ;
  // FP: "5 -> 6;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  wlnode_rup = ((0) + roundup(((*((volatile index_type *) (in_wl).dindex)) - (0)), (blockDim.x)));
  for (index_type wlnode = 0 + tid; wlnode < wlnode_rup; wlnode += nthreads)
  {
    int node;
    bool pop;
    multiple_sum<2, index_type> _np_mps;
    multiple_sum<2, index_type> _np_mps_total;
    // FP: "6 -> 7;
    // FP: "7 -> 8;
    pop = (in_wl).pop_id(wlnode, node);
    // FP: "8 -> 9;
    struct NPInspector1 _np = {0,0,0,0,0,0};
    // FP: "9 -> 10;
    if (pop)
    {
      _np.size = (graph).getOutDegree(node);
      _np.start = (graph).getFirstEdge(node);
    }
    // FP: "12 -> 13;
    // FP: "13 -> 14;
    _np_mps.el[0] = _np.size >= _NP_CROSSOVER_WP ? _np.size : 0;
    _np_mps.el[1] = _np.size < _NP_CROSSOVER_WP ? _np.size : 0;
    // FP: "14 -> 15;
    BlockScan(nps.temp_storage).ExclusiveSum(_np_mps, _np_mps, _np_mps_total);
    // FP: "15 -> 16;
    if (threadIdx.x == 0)
    {
      nps.tb.owner = MAX_TB_SIZE + 1;
    }
    // FP: "18 -> 19;
    __syncthreads();
    // FP: "19 -> 20;
    while (true)
    {
      // FP: "20 -> 21;
      if (_np.size >= _NP_CROSSOVER_TB)
      {
        nps.tb.owner = threadIdx.x;
      }
      // FP: "23 -> 24;
      __syncthreads();
      // FP: "24 -> 25;
      if (nps.tb.owner == MAX_TB_SIZE + 1)
      {
        // FP: "25 -> 26;
        __syncthreads();
        // FP: "26 -> 27;
        break;
      }
      // FP: "28 -> 29;
      if (nps.tb.owner == threadIdx.x)
      {
        nps.tb.start = _np.start;
        nps.tb.size = _np.size;
        nps.tb.src = threadIdx.x;
        _np.start = 0;
        _np.size = 0;
      }
      // FP: "31 -> 32;
      __syncthreads();
      // FP: "32 -> 33;
      int ns = nps.tb.start;
      int ne = nps.tb.size;
      // FP: "33 -> 34;
      if (nps.tb.src == threadIdx.x)
      {
        nps.tb.owner = MAX_TB_SIZE + 1;
      }
      // FP: "36 -> 37;
      for (int _np_j = threadIdx.x; _np_j < ne; _np_j += BLKSIZE)
      {
        index_type edge;
        edge = ns +_np_j;
        {
          index_type dst;
          dst = graph.getAbsDestination(edge);
          if (graph.node_data[dst] == INF)
          {
            index_type _start_23;
            graph.node_data[dst] = LEVEL;
            _start_23 = (out_wl).setup_push_warp_one();;
            (out_wl).do_push(_start_23, 0, dst);
          }
        }
      }
      // FP: "48 -> 49;
      __syncthreads();
    }
    // FP: "50 -> 51;

    // FP: "51 -> 52;
    {
      const int warpid = threadIdx.x / 32;
      // FP: "52 -> 53;
      const int _np_laneid = cub::LaneId();
      // FP: "53 -> 54;
      while (__any(_np.size >= _NP_CROSSOVER_WP && _np.size < _NP_CROSSOVER_TB))
      {
        if (_np.size >= _NP_CROSSOVER_WP && _np.size < _NP_CROSSOVER_TB)
        {
          nps.warp.owner[warpid] = _np_laneid;
        }
        if (nps.warp.owner[warpid] == _np_laneid)
        {
          nps.warp.start[warpid] = _np.start;
          nps.warp.size[warpid] = _np.size;

          _np.start = 0;
          _np.size = 0;
        }
        index_type _np_w_start = nps.warp.start[warpid];
        index_type _np_w_size = nps.warp.size[warpid];
        for (int _np_ii = _np_laneid; _np_ii < _np_w_size; _np_ii += 32)
        {
          index_type edge;
          edge = _np_w_start +_np_ii;
          {
            index_type dst;
            dst = graph.getAbsDestination(edge);
            if (graph.node_data[dst] == INF)
            {
              index_type _start_23;
              graph.node_data[dst] = LEVEL;
              _start_23 = (out_wl).setup_push_warp_one();;
              (out_wl).do_push(_start_23, 0, dst);
            }
          }
        }
      }
      // FP: "74 -> 75;
      __syncthreads();
      // FP: "75 -> 76;
    }

    // FP: "76 -> 77;
    __syncthreads();
    // FP: "77 -> 78;
    _np.total = _np_mps_total.el[1];
    _np.offset = _np_mps.el[1];
    // FP: "78 -> 79;
    while (_np.work())
    {
      // FP: "79 -> 80;
      int _np_i =0;
      // FP: "80 -> 81;
      _np.inspect(nps.fg.itvalue, ITSIZE);
      // FP: "81 -> 82;
      __syncthreads();
      // FP: "82 -> 83;

      // FP: "83 -> 84;
      for (_np_i = threadIdx.x; _np_i < ITSIZE && _np.valid(_np_i); _np_i += BLKSIZE)
      {
        index_type edge;
        edge= nps.fg.itvalue[_np_i];
        {
          index_type dst;
          dst = graph.getAbsDestination(edge);
          if (graph.node_data[dst] == INF)
          {
            index_type _start_23;
            graph.node_data[dst] = LEVEL;
            _start_23 = (out_wl).setup_push_warp_one();;
            (out_wl).do_push(_start_23, 0, dst);
          }
        }
      }
      // FP: "95 -> 96;
      _np.execute_round_done(ITSIZE);
      // FP: "96 -> 97;
      __syncthreads();
    }
  }
  // FP: "99 -> 100;
}
__global__ void bfs_kernel(CSRGraph graph, int LEVEL, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_bfs_kernel;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  bfs_kernel_dev(graph, LEVEL, in_wl, out_wl);
  // FP: "2 -> 3;
}
void gg_main_pipe_1(CSRGraph& gg, int& LEVEL, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  // FP: "1 -> 2;
  while (pipe.in_wl().nitems())
  {
    pipe.out_wl().will_write();
    bfs_kernel <<<blocks, __tb_bfs_kernel>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    pipe.advance2();
    LEVEL++;
  }
  // FP: "4 -> 5;
}
__global__ void __launch_bounds__(__tb_gg_main_pipe_1_gpu_gb) gg_main_pipe_1_gpu_gb(CSRGraph gg, int LEVEL, PipeContextT<Worklist2> pipe, int* cl_LEVEL, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_gg_main_pipe_1_gpu_gb;
  // FP: "1 -> 2;
  LEVEL = *cl_LEVEL;
  // FP: "2 -> 3;
  while (pipe.in_wl().nitems())
  {
    if (tid == 0)
      pipe.in_wl().reset_next_slot();
    bfs_kernel_dev (gg, LEVEL, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    gb.Sync();
    pipe.advance2();
    LEVEL++;
  }
  // FP: "5 -> 6;
  gb.Sync();
  // FP: "6 -> 7;
  if (tid == 0)
  {
    *cl_LEVEL = LEVEL;
  }
  // FP: "9 -> 10;
}
__global__ void gg_main_pipe_1_gpu(CSRGraph gg, int LEVEL, PipeContextT<Worklist2> pipe, dim3 blocks, dim3 threads, int* cl_LEVEL)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_one;
  // FP: "1 -> 2;
  LEVEL = *cl_LEVEL;
  // FP: "2 -> 3;
  while (pipe.in_wl().nitems())
  {
    bfs_kernel <<<blocks, __tb_bfs_kernel>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    cudaDeviceSynchronize();
    pipe.advance2();
    LEVEL++;
  }
  // FP: "5 -> 6;
  if (tid == 0)
  {
    *cl_LEVEL = LEVEL;
  }
  // FP: "8 -> 9;
}
void gg_main_pipe_1_wrapper(CSRGraph& gg, int& LEVEL, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  static GlobalBarrierLifetime gg_main_pipe_1_gpu_gb_barrier;
  static bool gg_main_pipe_1_gpu_gb_barrier_inited;
  // FP: "1 -> 2;
  static const size_t gg_main_pipe_1_gpu_gb_residency = maximum_residency(gg_main_pipe_1_gpu_gb, __tb_gg_main_pipe_1_gpu_gb, 0);
  static const size_t gg_main_pipe_1_gpu_gb_blocks = GG_MIN(blocks.x, ggc_get_nSM() * gg_main_pipe_1_gpu_gb_residency);
  if(!gg_main_pipe_1_gpu_gb_barrier_inited) { gg_main_pipe_1_gpu_gb_barrier.Setup(gg_main_pipe_1_gpu_gb_blocks); gg_main_pipe_1_gpu_gb_barrier_inited = true;};
  // FP: "2 -> 3;
  if (false)
  {
    gg_main_pipe_1(gg,LEVEL,pipe,blocks,threads);
  }
  else
  {
    int* cl_LEVEL;
    check_cuda(cudaMalloc(&cl_LEVEL, sizeof(int) * 1));
    check_cuda(cudaMemcpy(cl_LEVEL, &LEVEL, sizeof(int) * 1, cudaMemcpyHostToDevice));

    // gg_main_pipe_1_gpu<<<1,1>>>(gg,LEVEL,pipe,blocks,threads,cl_LEVEL);
    gg_main_pipe_1_gpu_gb<<<gg_main_pipe_1_gpu_gb_blocks, __tb_gg_main_pipe_1_gpu_gb>>>(gg,LEVEL,pipe,cl_LEVEL, gg_main_pipe_1_gpu_gb_barrier);
    check_cuda(cudaMemcpy(&LEVEL, cl_LEVEL, sizeof(int) * 1, cudaMemcpyDeviceToHost));
    check_cuda(cudaFree(cl_LEVEL));
  }
  // FP: "5 -> 6;
}
void gg_main(CSRGraph& hg, CSRGraph& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  PipeContextT<Worklist2> wl;
  // FP: "1 -> 2;
  bfs_init <<<blocks, threads>>>(gg, 0);
  // FP: "2 -> 3;
  int LEVEL = 1;
  // FP: "3 -> 4;
  wl = PipeContextT<Worklist2>(gg.nnodes);
  wl.in_wl().wl[0] = 0;
  wl.in_wl().update_gpu(1);
  gg_main_pipe_1_wrapper(gg,LEVEL,wl,blocks,threads);
  wl.free();
  // FP: "6 -> 7;
}