/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=True $ backoff_blocking_factor=4 $ parcomb=True $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=True $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
extern unsigned long DISCOUNT_TIME_NS;
static const int __tb_prep_edge_src = TB_SIZE;
static const int __tb_one = 1;
static const int __tb_gg_main_pipe_4_gpu_gb = 256;
static const int __tb_gg_main_pipe_3_gpu_gb = 256;
__global__ void init(CSRGraph graph)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = node;
  }
  // FP: "4 -> 5;
}
__global__ void prep_edge_src(CSRGraph graph, index_type * edge_src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_prep_edge_src;
  index_type node_end;
  index_type node_rup;
  // FP: "1 -> 2;
  const int _NP_CROSSOVER_WP = 32;
  const int _NP_CROSSOVER_TB = __kernel_tb_size;
  // FP: "2 -> 3;
  const int BLKSIZE = __kernel_tb_size;
  const int ITSIZE = BLKSIZE * 1;
  // FP: "3 -> 4;

  typedef cub::BlockScan<multiple_sum<2, index_type>, BLKSIZE> BlockScan;
  typedef union np_shared<BlockScan::TempStorage, index_type, struct tb_np, struct warp_np<__kernel_tb_size/32>, struct fg_np<ITSIZE> > npsTy;

  // FP: "4 -> 5;
  __shared__ npsTy nps ;
  // FP: "5 -> 6;
  node_end = (graph).nnodes;
  node_rup = ((0) + roundup((((graph).nnodes) - (0)), (blockDim.x)));
  for (index_type node = 0 + tid; node < node_rup; node += nthreads)
  {
    bool pop;
    multiple_sum<2, index_type> _np_mps;
    multiple_sum<2, index_type> _np_mps_total;
    // FP: "6 -> 7;
    // FP: "7 -> 8;
    pop = graph.valid_node(node);;
    // FP: "8 -> 9;
    struct NPInspector1 _np = {0,0,0,0,0,0};
    // FP: "9 -> 10;
    __shared__ struct { index_type node; } _np_closure [TB_SIZE];
    // FP: "10 -> 11;
    _np_closure[threadIdx.x].node = node;
    // FP: "11 -> 12;
    if (pop)
    {
      _np.size = (graph).getOutDegree(node);
      _np.start = (graph).getFirstEdge(node);
    }
    // FP: "14 -> 15;
    // FP: "15 -> 16;
    _np_mps.el[0] = _np.size >= _NP_CROSSOVER_WP ? _np.size : 0;
    _np_mps.el[1] = _np.size < _NP_CROSSOVER_WP ? _np.size : 0;
    // FP: "16 -> 17;
    BlockScan(nps.temp_storage).ExclusiveSum(_np_mps, _np_mps, _np_mps_total);
    // FP: "17 -> 18;
    if (threadIdx.x == 0)
    {
      nps.tb.owner = MAX_TB_SIZE + 1;
    }
    // FP: "20 -> 21;
    __syncthreads();
    // FP: "21 -> 22;
    while (true)
    {
      // FP: "22 -> 23;
      if (_np.size >= _NP_CROSSOVER_TB)
      {
        nps.tb.owner = threadIdx.x;
      }
      // FP: "25 -> 26;
      __syncthreads();
      // FP: "26 -> 27;
      if (nps.tb.owner == MAX_TB_SIZE + 1)
      {
        // FP: "27 -> 28;
        __syncthreads();
        // FP: "28 -> 29;
        break;
      }
      // FP: "30 -> 31;
      if (nps.tb.owner == threadIdx.x)
      {
        nps.tb.start = _np.start;
        nps.tb.size = _np.size;
        nps.tb.src = threadIdx.x;
        _np.start = 0;
        _np.size = 0;
      }
      // FP: "33 -> 34;
      __syncthreads();
      // FP: "34 -> 35;
      int ns = nps.tb.start;
      int ne = nps.tb.size;
      // FP: "35 -> 36;
      if (nps.tb.src == threadIdx.x)
      {
        nps.tb.owner = MAX_TB_SIZE + 1;
      }
      // FP: "38 -> 39;
      assert(nps.tb.src < __kernel_tb_size);
      node = _np_closure[nps.tb.src].node;
      // FP: "39 -> 40;
      for (int _np_j = threadIdx.x; _np_j < ne; _np_j += BLKSIZE)
      {
        index_type edge;
        edge = ns +_np_j;
        {
          edge_src[edge] = node;
        }
      }
      // FP: "44 -> 45;
      __syncthreads();
    }
    // FP: "46 -> 47;

    // FP: "47 -> 48;
    {
      const int warpid = threadIdx.x / 32;
      // FP: "48 -> 49;
      const int _np_laneid = cub::LaneId();
      // FP: "49 -> 50;
      while (__any(_np.size >= _NP_CROSSOVER_WP && _np.size < _NP_CROSSOVER_TB))
      {
        if (_np.size >= _NP_CROSSOVER_WP && _np.size < _NP_CROSSOVER_TB)
        {
          nps.warp.owner[warpid] = _np_laneid;
        }
        if (nps.warp.owner[warpid] == _np_laneid)
        {
          nps.warp.start[warpid] = _np.start;
          nps.warp.size[warpid] = _np.size;
          nps.warp.src[warpid] = threadIdx.x;
          _np.start = 0;
          _np.size = 0;
        }
        index_type _np_w_start = nps.warp.start[warpid];
        index_type _np_w_size = nps.warp.size[warpid];
        assert(nps.warp.src[warpid] < __kernel_tb_size);
        node = _np_closure[nps.warp.src[warpid]].node;
        for (int _np_ii = _np_laneid; _np_ii < _np_w_size; _np_ii += 32)
        {
          index_type edge;
          edge = _np_w_start +_np_ii;
          {
            edge_src[edge] = node;
          }
        }
      }
      // FP: "64 -> 65;
      __syncthreads();
      // FP: "65 -> 66;
    }

    // FP: "66 -> 67;
    __syncthreads();
    // FP: "67 -> 68;
    _np.total = _np_mps_total.el[1];
    _np.offset = _np_mps.el[1];
    // FP: "68 -> 69;
    while (_np.work())
    {
      // FP: "69 -> 70;
      int _np_i =0;
      // FP: "70 -> 71;
      _np.inspect2(nps.fg.itvalue, nps.fg.src, ITSIZE, threadIdx.x);
      // FP: "71 -> 72;
      __syncthreads();
      // FP: "72 -> 73;

      _np_i = threadIdx.x;
      // FP: "73 -> 74;
      if (_np.valid(_np_i))
      {
        index_type edge;
        assert(nps.fg.src[_np_i] < __kernel_tb_size);
        node = _np_closure[nps.fg.src[_np_i]].node;
        edge= nps.fg.itvalue[_np_i];
        {
          edge_src[edge] = node;
        }
      }
      // FP: "79 -> 80;
      _np.execute_round_done(ITSIZE);
      // FP: "80 -> 81;
      __syncthreads();
    }
    // FP: "82 -> 83;
    assert(threadIdx.x < __kernel_tb_size);
    node = _np_closure[threadIdx.x].node;
  }
  // FP: "84 -> 85;
}
__global__ void hook_init(CSRGraph graph, index_type * edge_src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  int edge_end;
  // FP: "1 -> 2;
  edge_end = graph.nedges;
  for (int edge = 0 + tid; edge < edge_end; edge += nthreads)
  {
    index_type x = edge_src[edge];
    index_type y = graph.getAbsDestination(edge);
    index_type mx = x > y ? x : y;
    index_type mn = x > y ? y : x;
    graph.node_data[mx] = mn;
  }
  // FP: "6 -> 7;
}
__global__ void hook_high_to_low(CSRGraph graph, const __restrict__ index_type * edge_src, char * marks, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  typedef cub::BlockReduce<int, TB_SIZE> _br;
  __shared__ _br::TempStorage _ts;
  ret_val.thread_entry();
  int edge_end;
  // FP: "1 -> 2;
  edge_end = graph.nedges;
  for (int edge = 0 + tid; edge < edge_end; edge += nthreads)
  {
    if (!marks[edge])
    {
      index_type u = edge_src[edge];
      index_type v = graph.getAbsDestination(edge);
      node_data_type p_u = graph.node_data[u];
      node_data_type p_v = graph.node_data[v];
      index_type mx = p_u > p_v ? p_u : p_v;
      index_type mn = p_u > p_v ? p_v : p_u;
      if (mx == mn)
      {
        marks[edge] = 1;
      }
      else
      {
        graph.node_data[mn] = mx;
        ret_val.do_return(true);
        continue;
      }
    }
  }
  ret_val.thread_exit<_br>(_ts);
}
__global__ void hook_low_to_high(CSRGraph graph, index_type * edge_src, char * marks, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  typedef cub::BlockReduce<int, TB_SIZE> _br;
  __shared__ _br::TempStorage _ts;
  ret_val.thread_entry();
  int edge_end;
  // FP: "1 -> 2;
  edge_end = graph.nedges;
  for (int edge = 0 + tid; edge < edge_end; edge += nthreads)
  {
    if (!marks[edge])
    {
      index_type u = edge_src[edge];
      index_type v = graph.getAbsDestination(edge);
      node_data_type p_u = graph.node_data[u];
      node_data_type p_v = graph.node_data[v];
      index_type mx = p_u > p_v ? p_u : p_v;
      index_type mn = p_u > p_v ? p_v : p_u;
      if (mx == mn)
      {
        marks[edge] = 1;
      }
      else
      {
        graph.node_data[mx] = mn;
        ret_val.do_return(true);
        continue;
      }
    }
  }
  ret_val.thread_exit<_br>(_ts);
}
__device__ void p_jump_dev(CSRGraph graph, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  typedef cub::BlockReduce<int, TB_SIZE> _br;
  __shared__ _br::TempStorage _ts;
  ret_val.thread_entry();
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    node_data_type p_u = graph.node_data[node];
    node_data_type p_v = graph.node_data[p_u];
    if (p_u != p_v)
    {
      graph.node_data[node] = p_v;
      ret_val.do_return(true);
      continue;
    }
  }
  ret_val.thread_exit<_br>(_ts);
}
__global__ void p_jump(CSRGraph graph, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  // FP: "1 -> 2;
  p_jump_dev(graph, ret_val);
  // FP: "2 -> 3;
}
__global__ void identify_roots(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (graph.node_data[node] == node)
    {
      index_type _start_73;
      _start_73 = (out_wl).setup_push_warp_one();;
      (out_wl).do_push(_start_73, 0, node);
    }
  }
  // FP: "8 -> 9;
}
__device__ void p_jump_roots_dev(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  typedef cub::BlockReduce<int, TB_SIZE> _br;
  __shared__ _br::TempStorage _ts;
  ret_val.thread_entry();
  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    pop = (in_wl).pop_id(wlnode, node);
    node_data_type p_u = graph.node_data[node];
    node_data_type p_v = graph.node_data[p_u];
    if (p_u != p_v)
    {
      graph.node_data[node] = p_v;
      ret_val.do_return(true);
      continue;
    }
  }
  ret_val.thread_exit<_br>(_ts);
}
__global__ void p_jump_roots(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  p_jump_roots_dev(graph, in_wl, out_wl, ret_val);
  // FP: "2 -> 3;
}
__global__ void p_jump_leaves(CSRGraph graph)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    node_data_type p_u = graph.node_data[node];
    node_data_type p_v = graph.node_data[p_u];
    if (p_u != p_v)
    {
      graph.node_data[node] = p_v;
    }
  }
  // FP: "7 -> 8;
}
__global__ void count_components(CSRGraph graph, int * count)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (node == graph.node_data[node])
    {
      atomicAdd(count, 1);
    }
  }
  // FP: "6 -> 7;
}
void gg_main_pipe_4(CSRGraphTy& gg, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  bool loopc = false;
  // FP: "1 -> 2;
  do
  {
    Shared<int> retval = Shared<int>(1);
    Any _rv;
    *(retval.cpu_wr_ptr()) = 0;
    _rv.rv = retval.gpu_wr_ptr();
    pipe.out_wl().will_write();
    p_jump_roots <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl(), _rv);
    loopc = *(retval.cpu_rd_ptr()) > 0;
    // FP: "2 -> 2;
  }
  while (loopc);
  // FP: "3 -> 4;
}
__global__ void __launch_bounds__(__tb_gg_main_pipe_4_gpu_gb) gg_main_pipe_4_gpu_gb(CSRGraphTy gg, PipeContextT<Worklist2> pipe, int* retval, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  bool loopc = false;
  // FP: "1 -> 2;
  do
  {
    Any _rv;
    *retval = 0;
    _rv.rv = retval;
    gb.Sync();
    if (tid == 0)
      pipe.in_wl().reset_next_slot();
    p_jump_roots_dev (gg, pipe.in_wl(), pipe.out_wl(), _rv);
    _rv.local = *retval;
    gb.Sync();
    loopc = *retval > 0;
    // FP: "2 -> 2;
    gb.Sync();
  }
  while (loopc);
  // FP: "3 -> 4;
  gb.Sync();
  // FP: "4 -> 5;
  if (tid == 0)
  {
    pipe.save();
  }
  // FP: "7 -> 8;
}
__global__ void gg_main_pipe_4_gpu(CSRGraphTy gg, PipeContextT<Worklist2> pipe, dim3 blocks, dim3 threads, int* retval)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_one;
  bool loopc = false;
  // FP: "1 -> 2;
  do
  {
    Any _rv;
    *retval = 0;
    _rv.rv = retval;
    p_jump_roots <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl(), _rv);
    cudaDeviceSynchronize();
    loopc = *retval > 0;
    // FP: "2 -> 2;
  }
  while (loopc);
  // FP: "3 -> 4;
  if (tid == 0)
  {
    pipe.save();
  }
  // FP: "6 -> 7;
}
void gg_main_pipe_4_wrapper(CSRGraphTy& gg, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  static GlobalBarrierLifetime gg_main_pipe_4_gpu_gb_barrier;
  static bool gg_main_pipe_4_gpu_gb_barrier_inited;
  // FP: "1 -> 2;
  static const size_t gg_main_pipe_4_gpu_gb_residency = maximum_residency(gg_main_pipe_4_gpu_gb, __tb_gg_main_pipe_4_gpu_gb, 0);
  static const size_t gg_main_pipe_4_gpu_gb_blocks = GG_MIN(blocks.x, ggc_get_nSM() * gg_main_pipe_4_gpu_gb_residency);
  if(!gg_main_pipe_4_gpu_gb_barrier_inited) { gg_main_pipe_4_gpu_gb_barrier.Setup(gg_main_pipe_4_gpu_gb_blocks); gg_main_pipe_4_gpu_gb_barrier_inited = true;};
  // FP: "2 -> 3;
  Shared<int> retval (1);
  // FP: "3 -> 4;
  if (false)
  {
    gg_main_pipe_4(gg,pipe,blocks,threads);
  }
  else
  {
    pipe.prep();

    // gg_main_pipe_4_gpu<<<1,1>>>(gg,pipe,blocks,threads,retval.gpu_wr_ptr());
    gg_main_pipe_4_gpu_gb<<<gg_main_pipe_4_gpu_gb_blocks, __tb_gg_main_pipe_4_gpu_gb>>>(gg,pipe,retval.gpu_wr_ptr(), gg_main_pipe_4_gpu_gb_barrier);
    pipe.restore();
  }
  // FP: "6 -> 7;
}
void gg_main_pipe_3(CSRGraphTy& gg, dim3& blocks, dim3& threads)
{
  bool loopc = false;
  // FP: "1 -> 2;
  do
  {
    Shared<int> retval = Shared<int>(1);
    Any _rv;
    *(retval.cpu_wr_ptr()) = 0;
    _rv.rv = retval.gpu_wr_ptr();
    p_jump <<<blocks, threads>>>(gg, _rv);
    loopc = *(retval.cpu_rd_ptr()) > 0;
    // FP: "2 -> 2;
  }
  while (loopc);
  // FP: "3 -> 4;
}
__global__ void __launch_bounds__(__tb_gg_main_pipe_3_gpu_gb) gg_main_pipe_3_gpu_gb(CSRGraphTy gg, int* retval, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  bool loopc = false;
  // FP: "1 -> 2;
  do
  {
    Any _rv;
    *retval = 0;
    _rv.rv = retval;
    gb.Sync();
    p_jump_dev (gg, _rv);
    _rv.local = *retval;
    gb.Sync();
    loopc = *retval > 0;
    // FP: "2 -> 2;
    gb.Sync();
  }
  while (loopc);
  // FP: "3 -> 4;
  gb.Sync();
  // FP: "4 -> 5;
}
__global__ void gg_main_pipe_3_gpu(CSRGraphTy gg, dim3 blocks, dim3 threads, int* retval)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_one;
  bool loopc = false;
  // FP: "1 -> 2;
  do
  {
    Any _rv;
    *retval = 0;
    _rv.rv = retval;
    p_jump <<<blocks, threads>>>(gg, _rv);
    cudaDeviceSynchronize();
    loopc = *retval > 0;
    // FP: "2 -> 2;
  }
  while (loopc);
  // FP: "3 -> 4;
}
void gg_main_pipe_3_wrapper(CSRGraphTy& gg, dim3& blocks, dim3& threads)
{
  static GlobalBarrierLifetime gg_main_pipe_3_gpu_gb_barrier;
  static bool gg_main_pipe_3_gpu_gb_barrier_inited;
  // FP: "1 -> 2;
  static const size_t gg_main_pipe_3_gpu_gb_residency = maximum_residency(gg_main_pipe_3_gpu_gb, __tb_gg_main_pipe_3_gpu_gb, 0);
  static const size_t gg_main_pipe_3_gpu_gb_blocks = GG_MIN(blocks.x, ggc_get_nSM() * gg_main_pipe_3_gpu_gb_residency);
  if(!gg_main_pipe_3_gpu_gb_barrier_inited) { gg_main_pipe_3_gpu_gb_barrier.Setup(gg_main_pipe_3_gpu_gb_blocks); gg_main_pipe_3_gpu_gb_barrier_inited = true;};
  // FP: "2 -> 3;
  Shared<int> retval (1);
  // FP: "3 -> 4;
  if (false)
  {
    gg_main_pipe_3(gg,blocks,threads);
  }
  else
  {

    // gg_main_pipe_3_gpu<<<1,1>>>(gg,blocks,threads,retval.gpu_wr_ptr());
    gg_main_pipe_3_gpu_gb<<<gg_main_pipe_3_gpu_gb_blocks, __tb_gg_main_pipe_3_gpu_gb>>>(gg,retval.gpu_wr_ptr(), gg_main_pipe_3_gpu_gb_barrier);
  }
  // FP: "6 -> 7;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  int edge_blocks;
  int node_blocks;
  cudaEvent_t start;
  cudaEvent_t stop;
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  int it_hk = 1;
  // FP: "2 -> 3;
  Shared<index_type> edge_src (gg.nedges);
  // FP: "3 -> 4;
  Shared<char> edge_marks (gg.nedges);
  // FP: "4 -> 5;
  bool flag = false;
  // FP: "5 -> 6;
  // FP: "6 -> 7;
  edge_blocks = hg.nedges / TB_SIZE + 1;
  node_blocks = hg.nnodes / TB_SIZE + 1;
  // FP: "7 -> 8;
  edge_marks.zero_gpu();
  // FP: "8 -> 9;
  // FP: "9 -> 10;
  check_cuda(cudaEventCreate(&start));
  check_cuda(cudaEventCreate(&stop));
  // FP: "10 -> 11;
  check_cuda(cudaEventRecord(start));
  // FP: "11 -> 12;
  prep_edge_src <<<node_blocks, __tb_prep_edge_src>>>(gg, edge_src.gpu_wr_ptr());
  // FP: "12 -> 13;
  check_cuda(cudaEventRecord(stop));
  // FP: "13 -> 14;
  init <<<node_blocks, threads>>>(gg);
  // FP: "14 -> 15;
  hook_init <<<edge_blocks, threads>>>(gg, edge_src.gpu_rd_ptr());
  // FP: "15 -> 16;
  gg_main_pipe_3_wrapper(gg,blocks,threads);
  // FP: "17 -> 18;
  pipe = PipeContextT<Worklist2>(gg.nnodes);
  {
    {
      do
      {
        pipe.out_wl().will_write();
        identify_roots <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        if (it_hk != 0)
        {
          Shared<int> retval = Shared<int>(1);
          Any _rv;
          *(retval.cpu_wr_ptr()) = 0;
          _rv.rv = retval.gpu_wr_ptr();
          hook_low_to_high <<<edge_blocks, threads>>>(gg, edge_src.gpu_rd_ptr(), edge_marks.gpu_wr_ptr(), _rv);
          flag = *(retval.cpu_rd_ptr());
          it_hk = (it_hk + 1) % 4;
        }
        else
        {
          Shared<int> retval = Shared<int>(1);
          Any _rv;
          *(retval.cpu_wr_ptr()) = 0;
          _rv.rv = retval.gpu_wr_ptr();
          hook_high_to_low <<<edge_blocks, threads>>>(gg, edge_src.gpu_rd_ptr(), edge_marks.gpu_wr_ptr(), _rv);
          flag = *(retval.cpu_rd_ptr());
        }
        if (!flag)
        {
          break;
        }
        gg_main_pipe_4_wrapper(gg,pipe,blocks,threads);
        p_jump_leaves <<<node_blocks, threads>>>(gg);
      }
      while (flag);
      // FP: "36 -> 37;
    }
  }
  pipe.free();
  // FP: "18 -> 19;
  printf("iterations: %d\n", it_hk);
  // FP: "37 -> 38;
  Shared<int> count (1);
  // FP: "38 -> 39;
  *(count.cpu_wr_ptr()) = 0;
  // FP: "39 -> 40;
  count_components <<<blocks, threads>>>(gg, count.gpu_wr_ptr());
  // FP: "40 -> 41;
  printf("components: %d\n", *(count.cpu_rd_ptr()));
  // FP: "41 -> 42;
  float ms =0;
  // FP: "42 -> 43;
  check_cuda(cudaEventElapsedTime(&ms, start, stop));
  DISCOUNT_TIME_NS = (int) (ms * 1000000);
  printf("prep_edge_src: %llu ns\n", DISCOUNT_TIME_NS);
  // FP: "43 -> 44;
}