/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=True $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
#include <curand.h>
#define UNMARKED 0
#define MARKED 1
#define NON_INDEPENDENT 2
#define NON_MAXIMAL 3
#define SEED1 0x12345678LL
#define SEED2 0xabbdef12LL
#define SEED3 0xcafe1234LL
#define SEED4 0x09832516LL
static const int __tb_one = 1;
static const int __tb_gg_main_pipe_1_gpu_gb = 256;
__global__ void gen_prio_gpu(CSRGraph graph, unsigned int * prio, unsigned int x, unsigned int y, unsigned int z, unsigned int w)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  x ^= tid;
  y ^= tid;
  z ^= tid;
  w ^= tid;
  assert(!(x == 0 && y == 0 && z == 0 && w == 0));
  // FP: "2 -> 3;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    unsigned int t;
    t = x ^ (x << 11);
    x = y;
    y = z;
    z = w;
    w = w ^ (w >> 19) ^ t ^ (t >> 8);
    prio[node] = w;
  }
  // FP: "6 -> 7;
}
void gen_prio(CSRGraph graph, unsigned int * prio)
{
  curandGenerator_t gen;
  // FP: "1 -> 2;
  // FP: "2 -> 3;
  check_rv(curandCreateGenerator(&gen, CURAND_RNG_PSEUDO_MT19937), CURAND_STATUS_SUCCESS);
  check_rv(curandSetPseudoRandomGeneratorSeed(gen, SEED1), CURAND_STATUS_SUCCESS);
  check_rv(curandSetGeneratorOrdering (gen, CURAND_ORDERING_PSEUDO_BEST), CURAND_STATUS_SUCCESS);
  // FP: "3 -> 4;
}
__device__ void init_wl_dev(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    (out_wl).push(node);
  }
  // FP: "4 -> 5;
}
__global__ void init_wl(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  init_wl_dev(graph, in_wl, out_wl);
  // FP: "2 -> 3;
}
__device__ void mark_nodes_dev(CSRGraph graph, const unsigned int * __restrict__ prio, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    int max_prio = prio[node];
    int max_prio_node = node;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst = graph.getAbsDestination(edge);
      if (dst != node && graph.node_data[dst] != NON_INDEPENDENT && prio[dst] >= max_prio)
      {
        if ((prio[dst] > max_prio) || dst > max_prio_node)
        {
          max_prio = prio[dst];
          max_prio_node = dst;
        }
      }
    }
    if (max_prio_node == node)
    {
      assert(graph.node_data[node] == UNMARKED);
      graph.node_data[node] = MARKED;
    }
  }
  // FP: "17 -> 18;
}
__global__ void mark_nodes(CSRGraph graph, const unsigned int * __restrict__ prio, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  mark_nodes_dev(graph, prio, in_wl, out_wl);
  // FP: "2 -> 3;
}
__device__ void drop_marked_nodes_and_nbors_dev(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    pop = (in_wl).pop_id(wlnode, node);
    bool drop = false;
    if (graph.node_data[node] == MARKED)
    {
      drop = true;
    }
    if (!drop)
    {
      index_type edge_end;
      edge_end = (graph).getFirstEdge((node) + 1);
      for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
      {
        index_type dst = graph.getAbsDestination(edge);
        if (graph.node_data[dst] == MARKED)
        {
          drop = true;
        }
      }
    }
    if (!drop)
    {
      (out_wl).push(node);
    }
    else
    {
      if (graph.node_data[node] == UNMARKED)
      {
        graph.node_data[node] = NON_INDEPENDENT;
      }
    }
  }
  // FP: "23 -> 24;
}
__global__ void drop_marked_nodes_and_nbors(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  drop_marked_nodes_and_nbors_dev(graph, in_wl, out_wl);
  // FP: "2 -> 3;
}
void gg_main_pipe_1(CSRGraphTy& gg, int& STEPS, Shared<unsigned int>& prio, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  // FP: "1 -> 2;
  {
    pipe.out_wl().will_write();
    init_wl <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    pipe.advance2();
    // FP: "3 -> 4;
    while (pipe.in_wl().nitems())
    {
      pipe.out_wl().will_write();
      mark_nodes <<<blocks, threads>>>(gg, prio.gpu_rd_ptr(), pipe.in_wl(), pipe.out_wl());
      pipe.out_wl().will_write();
      drop_marked_nodes_and_nbors <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      STEPS++;
    }
    // FP: "8 -> 9;
  }
  // FP: "2 -> 3;
}
__global__ void __launch_bounds__(__tb_gg_main_pipe_1_gpu_gb) gg_main_pipe_1_gpu_gb(CSRGraphTy gg, int STEPS, unsigned int* prio, PipeContextT<Worklist2> pipe, int* cl_STEPS, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  // FP: "1 -> 2;
  STEPS = *cl_STEPS;
  // FP: "2 -> 3;
  {
    if (tid == 0)
      pipe.in_wl().reset_next_slot();
    init_wl_dev (gg, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    gb.Sync();
    pipe.advance2();
    // FP: "4 -> 5;
    while (pipe.in_wl().nitems())
    {
      if (tid == 0)
        pipe.in_wl().reset_next_slot();
      mark_nodes_dev (gg, prio, pipe.in_wl(), pipe.out_wl());
      gb.Sync();
      if (tid == 0)
        pipe.in_wl().reset_next_slot();
      drop_marked_nodes_and_nbors_dev (gg, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      gb.Sync();
      pipe.advance2();
      STEPS++;
    }
    // FP: "9 -> 10;
  }
  // FP: "3 -> 4;
  gb.Sync();
  // FP: "10 -> 11;
  if (tid == 0)
  {
    *cl_STEPS = STEPS;
  }
  // FP: "13 -> 14;
}
__global__ void gg_main_pipe_1_gpu(CSRGraphTy gg, int STEPS, unsigned int* prio, PipeContextT<Worklist2> pipe, dim3 blocks, dim3 threads, int* cl_STEPS)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_one;
  // FP: "1 -> 2;
  STEPS = *cl_STEPS;
  // FP: "2 -> 3;
  {
    init_wl <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    cudaDeviceSynchronize();
    pipe.advance2();
    // FP: "4 -> 5;
    while (pipe.in_wl().nitems())
    {
      mark_nodes <<<blocks, threads>>>(gg, prio, pipe.in_wl(), pipe.out_wl());
      cudaDeviceSynchronize();
      drop_marked_nodes_and_nbors <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      cudaDeviceSynchronize();
      pipe.advance2();
      STEPS++;
    }
    // FP: "9 -> 10;
  }
  // FP: "3 -> 4;
  if (tid == 0)
  {
    *cl_STEPS = STEPS;
  }
  // FP: "12 -> 13;
}
void gg_main_pipe_1_wrapper(CSRGraphTy& gg, int& STEPS, Shared<unsigned int>& prio, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  static GlobalBarrierLifetime gg_main_pipe_1_gpu_gb_barrier;
  static bool gg_main_pipe_1_gpu_gb_barrier_inited;
  // FP: "1 -> 2;
  static const size_t gg_main_pipe_1_gpu_gb_residency = maximum_residency(gg_main_pipe_1_gpu_gb, __tb_gg_main_pipe_1_gpu_gb, 0);
  static const size_t gg_main_pipe_1_gpu_gb_blocks = GG_MIN(blocks.x, ggc_get_nSM() * gg_main_pipe_1_gpu_gb_residency);
  if(!gg_main_pipe_1_gpu_gb_barrier_inited) { gg_main_pipe_1_gpu_gb_barrier.Setup(gg_main_pipe_1_gpu_gb_blocks); gg_main_pipe_1_gpu_gb_barrier_inited = true;};
  // FP: "2 -> 3;
  if (false)
  {
    gg_main_pipe_1(gg,STEPS,prio,pipe,blocks,threads);
  }
  else
  {
    int* cl_STEPS;
    check_cuda(cudaMalloc(&cl_STEPS, sizeof(int) * 1));
    check_cuda(cudaMemcpy(cl_STEPS, &STEPS, sizeof(int) * 1, cudaMemcpyHostToDevice));

    // gg_main_pipe_1_gpu<<<1,1>>>(gg,STEPS,prio.gpu_wr_ptr(),pipe,blocks,threads,cl_STEPS);
    gg_main_pipe_1_gpu_gb<<<gg_main_pipe_1_gpu_gb_blocks, __tb_gg_main_pipe_1_gpu_gb>>>(gg,STEPS,prio.gpu_wr_ptr(),pipe,cl_STEPS, gg_main_pipe_1_gpu_gb_barrier);
    check_cuda(cudaMemcpy(&STEPS, cl_STEPS, sizeof(int) * 1, cudaMemcpyDeviceToHost));
    check_cuda(cudaFree(cl_STEPS));
  }
  // FP: "5 -> 6;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  Shared<unsigned int> prio (hg.nnodes);
  int STEPS = 0;
  // FP: "2 -> 3;
  ggc::Timer t ("random");
  // FP: "3 -> 4;
  t.start();
  // FP: "4 -> 5;
  gen_prio_gpu <<<blocks, threads>>>(gg, prio.gpu_wr_ptr(), SEED1, SEED2, SEED3, SEED4);
  // FP: "5 -> 6;
  cudaDeviceSynchronize();
  // FP: "6 -> 7;
  t.stop();
  printf("Random number generation took %llu ns\n", t.duration());
  // FP: "7 -> 8;
  pipe = PipeContextT<Worklist2>(gg.nnodes);
  gg_main_pipe_1_wrapper(gg,STEPS,prio,pipe,blocks,threads);
  pipe.free();
  // FP: "8 -> 9;
  printf("Total steps: %d\n", STEPS);
  // FP: "15 -> 16;
}