/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=True $ backoff_blocking_factor=4 $ parcomb=True $ np_schedulers=set(['wp', 'fg']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=8 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=True $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
#include "kernels/reduce.cuh"
typedef int edge_data_type;
typedef int node_data_type;
typedef float* gfloat_p;
float* P_CURR ;
float* P_NEXT ;
extern const float ALPHA = 0.85;
extern const float EPSILON = 0.01;
extern int MAX_ITERATIONS ;
static const int __tb_gg_main_pipe_1_gpu_gb = 256;
static const int __tb_one = 1;
static const int __tb_pagerank_main = TB_SIZE;
static const int __tb_remove_dups = TB_SIZE;
__global__ void init_1(CSRGraphTex graph, float * p_curr, float * residual)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    float update;
    index_type edge_end;
    p_curr[node] = 1.0 - ALPHA;
    update = 1.0/graph.getOutDegree(node);
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      dst = graph.getAbsDestination(edge);
      atomicAdd(residual + dst, update);
    }
  }
  // FP: "10 -> 11;
}
__device__ void init_2_dev(CSRGraphTex graph, float * residual, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type _start_22;
  index_type node_end;
  // FP: "1 -> 2;
  // FP: "2 -> 3;
  _start_22 = (out_wl).push_range((tid < ((graph).nnodes)) ? ((((graph).nnodes) - 1 - tid)/nthreads + 1) : 0);;
  // FP: "3 -> 4;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid, node_pos = 0; node < node_end; node_pos++, node += nthreads)
  {
    residual[node] *= (1.0 - ALPHA) * ALPHA;
    (out_wl).do_push(_start_22, node_pos, node);
  }
  // FP: "7 -> 8;
}
__global__ void init_2(CSRGraphTex graph, float * residual, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  init_2_dev(graph, residual, in_wl, out_wl);
  // FP: "2 -> 3;
}
__global__ void remove_dups(int * marks, Worklist2 in_wl, Worklist2 out_wl, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  index_type wlnode2_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    marks[node] = wlnode;
  }
  // FP: "6 -> 7;
  gb.Sync();
  // FP: "7 -> 8;
  wlnode2_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode2 = 0 + tid; wlnode2 < wlnode2_end; wlnode2 += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode2, node);
    if (marks[node] == wlnode2)
    {
      index_type _start_37;
      _start_37 = (out_wl).setup_push_warp_one();;
      (out_wl).do_push(_start_37, 0, node);
    }
  }
  // FP: "16 -> 17;
}
__device__ void pagerank_main_dev(CSRGraphTex graph, float * p_curr, float * residual, float * p_diff, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_pagerank_main;
  index_type wlnode_end;
  index_type wlnode_rup;
  // FP: "1 -> 2;
  const int _NP_CROSSOVER_WP = 32;
  const int _NP_CROSSOVER_TB = __kernel_tb_size;
  // FP: "2 -> 3;
  const int BLKSIZE = __kernel_tb_size;
  const int ITSIZE = BLKSIZE * 8;
  // FP: "3 -> 4;

  typedef cub::BlockScan<multiple_sum<2, index_type>, BLKSIZE> BlockScan;
  typedef union np_shared<BlockScan::TempStorage, index_type, struct empty_np, struct warp_np<__kernel_tb_size/32>, struct fg_np<ITSIZE> > npsTy;

  // FP: "4 -> 5;
  __shared__ npsTy nps ;
  // FP: "5 -> 6;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  wlnode_rup = ((0) + roundup(((*((volatile index_type *) (in_wl).dindex)) - (0)), (blockDim.x)));
  for (index_type wlnode = 0 + tid; wlnode < wlnode_rup; wlnode += nthreads)
  {
    int sdeg;
    float update;
    int node;
    bool pop;
    float res;
    multiple_sum<2, index_type> _np_mps;
    multiple_sum<2, index_type> _np_mps_total;
    // FP: "6 -> 7;
    // FP: "7 -> 8;
    // FP: "8 -> 9;
    pop = (in_wl).pop_id(wlnode, node);
    // FP: "9 -> 10;
    // FP: "10 -> 11;
    if (pop)
    {
      res =atomicExch(residual + node, 0);
      p_curr[node] += res;
      sdeg = graph.getOutDegree(node);
      update = res * ALPHA / sdeg;
    }
    // FP: "13 -> 14;
    struct NPInspector1 _np = {0,0,0,0,0,0};
    // FP: "14 -> 15;
    __shared__ struct { float update; } _np_closure [TB_SIZE];
    // FP: "15 -> 16;
    _np_closure[threadIdx.x].update = update;
    // FP: "16 -> 17;
    if (pop)
    {
      _np.size = (graph).getOutDegree(node);
      _np.start = (graph).getFirstEdge(node);
    }
    // FP: "19 -> 20;
    // FP: "20 -> 21;
    _np_mps.el[0] = _np.size >= _NP_CROSSOVER_WP ? _np.size : 0;
    _np_mps.el[1] = _np.size < _NP_CROSSOVER_WP ? _np.size : 0;
    // FP: "21 -> 22;
    BlockScan(nps.temp_storage).ExclusiveSum(_np_mps, _np_mps, _np_mps_total);
    // FP: "22 -> 23;
    if (threadIdx.x == 0)
    {
    }
    // FP: "24 -> 25;
    __syncthreads();
    // FP: "25 -> 26;
    {
      const int warpid = threadIdx.x / 32;
      // FP: "26 -> 27;
      const int _np_laneid = cub::LaneId();
      // FP: "27 -> 28;
      while (__any(_np.size >= _NP_CROSSOVER_WP))
      {
        if (_np.size >= _NP_CROSSOVER_WP)
        {
          nps.warp.owner[warpid] = _np_laneid;
        }
        if (nps.warp.owner[warpid] == _np_laneid)
        {
          nps.warp.start[warpid] = _np.start;
          nps.warp.size[warpid] = _np.size;
          nps.warp.src[warpid] = threadIdx.x;
          _np.start = 0;
          _np.size = 0;
        }
        index_type _np_w_start = nps.warp.start[warpid];
        index_type _np_w_size = nps.warp.size[warpid];
        assert(nps.warp.src[warpid] < __kernel_tb_size);
        update = _np_closure[nps.warp.src[warpid]].update;
        for (int _np_ii = _np_laneid; _np_ii < _np_w_size; _np_ii += 32)
        {
          index_type edge;
          edge = _np_w_start +_np_ii;
          {
            index_type dst;
            float prev;
            dst = graph.getAbsDestination(edge);
            prev = atomicAdd(residual + dst, update);
            if (prev + update > EPSILON && prev < EPSILON)
            {
              index_type _start_57;
              _start_57 = (out_wl).setup_push_warp_one();;
              (out_wl).do_push(_start_57, 0, dst);
            }
          }
        }
      }
      // FP: "48 -> 49;
      __syncthreads();
      // FP: "49 -> 50;
    }

    // FP: "50 -> 51;
    __syncthreads();
    // FP: "51 -> 52;
    _np.total = _np_mps_total.el[1];
    _np.offset = _np_mps.el[1];
    // FP: "52 -> 53;
    while (_np.work())
    {
      // FP: "53 -> 54;
      int _np_i =0;
      // FP: "54 -> 55;
      _np.inspect2(nps.fg.itvalue, nps.fg.src, ITSIZE, threadIdx.x);
      // FP: "55 -> 56;
      __syncthreads();
      // FP: "56 -> 57;

      // FP: "57 -> 58;
      for (_np_i = threadIdx.x; _np_i < ITSIZE && _np.valid(_np_i); _np_i += BLKSIZE)
      {
        index_type edge;
        assert(nps.fg.src[_np_i] < __kernel_tb_size);
        update = _np_closure[nps.fg.src[_np_i]].update;
        edge= nps.fg.itvalue[_np_i];
        {
          index_type dst;
          float prev;
          dst = graph.getAbsDestination(edge);
          prev = atomicAdd(residual + dst, update);
          if (prev + update > EPSILON && prev < EPSILON)
          {
            index_type _start_57;
            _start_57 = (out_wl).setup_push_warp_one();;
            (out_wl).do_push(_start_57, 0, dst);
          }
        }
      }
      // FP: "69 -> 70;
      _np.execute_round_done(ITSIZE);
      // FP: "70 -> 71;
      __syncthreads();
    }
    // FP: "72 -> 73;
    assert(threadIdx.x < __kernel_tb_size);
    update = _np_closure[threadIdx.x].update;
  }
  // FP: "74 -> 75;
}
__global__ void __launch_bounds__(TB_SIZE, 3) pagerank_main(CSRGraphTex graph, float * p_curr, float * residual, float * p_diff, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_pagerank_main;
  if (tid == 0)
    in_wl.reset_next_slot();

  // FP: "1 -> 2;
  pagerank_main_dev(graph, p_curr, residual, p_diff, in_wl, out_wl);
  // FP: "2 -> 3;
}
void gg_main_pipe_1(gfloat_p p2, gfloat_p p0, gfloat_p rp, int& iter, CSRGraphTex& gg, CSRGraphTex& hg, int MAX_ITERATIONS, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  // FP: "1 -> 2;
  {
    pipe.out_wl().will_write();
    init_2 <<<blocks, threads>>>(gg, rp, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    pipe.advance2();
    // FP: "3 -> 4;
    while (pipe.in_wl().nitems())
    {
      pipe.out_wl().will_write();
      pagerank_main <<<blocks, __tb_pagerank_main>>>(gg, p0, rp, p2, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      iter++;
      if (iter >= MAX_ITERATIONS)
      {
        break;
      }
    }
    // FP: "10 -> 11;
  }
  // FP: "2 -> 3;
}
__global__ void __launch_bounds__(__tb_gg_main_pipe_1_gpu_gb) gg_main_pipe_1_gpu_gb(gfloat_p p2, gfloat_p p0, gfloat_p rp, int iter, CSRGraphTex gg, CSRGraphTex hg, int MAX_ITERATIONS, PipeContextT<Worklist2> pipe, int* cl_iter, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_gg_main_pipe_1_gpu_gb;
  // FP: "1 -> 2;
  iter = *cl_iter;
  // FP: "2 -> 3;
  {
    if (tid == 0)
      pipe.in_wl().reset_next_slot();
    init_2_dev (gg, rp, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    gb.Sync();
    pipe.advance2();
    // FP: "4 -> 5;
    while (pipe.in_wl().nitems())
    {
      if (tid == 0)
        pipe.in_wl().reset_next_slot();
      pagerank_main_dev (gg, p0, rp, p2, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      gb.Sync();
      pipe.advance2();
      iter++;
      if (iter >= MAX_ITERATIONS)
      {
        break;
      }
    }
    // FP: "11 -> 12;
  }
  // FP: "3 -> 4;
  gb.Sync();
  // FP: "12 -> 13;
  if (tid == 0)
  {
    *cl_iter = iter;
  }
  // FP: "15 -> 16;
}
__global__ void gg_main_pipe_1_gpu(gfloat_p p2, gfloat_p p0, gfloat_p rp, int iter, CSRGraphTex gg, CSRGraphTex hg, int MAX_ITERATIONS, PipeContextT<Worklist2> pipe, dim3 blocks, dim3 threads, int* cl_iter)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_one;
  // FP: "1 -> 2;
  iter = *cl_iter;
  // FP: "2 -> 3;
  {
    init_2 <<<blocks, threads>>>(gg, rp, pipe.in_wl(), pipe.out_wl());
    pipe.in_wl().swap_slots();
    cudaDeviceSynchronize();
    pipe.advance2();
    // FP: "4 -> 5;
    while (pipe.in_wl().nitems())
    {
      pagerank_main <<<blocks, __tb_pagerank_main>>>(gg, p0, rp, p2, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      cudaDeviceSynchronize();
      pipe.advance2();
      iter++;
      if (iter >= MAX_ITERATIONS)
      {
        break;
      }
    }
    // FP: "11 -> 12;
  }
  // FP: "3 -> 4;
  if (tid == 0)
  {
    *cl_iter = iter;
  }
  // FP: "14 -> 15;
}
void gg_main_pipe_1_wrapper(gfloat_p p2, gfloat_p p0, gfloat_p rp, int& iter, CSRGraphTex& gg, CSRGraphTex& hg, int MAX_ITERATIONS, PipeContextT<Worklist2>& pipe, dim3& blocks, dim3& threads)
{
  static GlobalBarrierLifetime gg_main_pipe_1_gpu_gb_barrier;
  static bool gg_main_pipe_1_gpu_gb_barrier_inited;
  // FP: "1 -> 2;
  static const size_t gg_main_pipe_1_gpu_gb_residency = maximum_residency(gg_main_pipe_1_gpu_gb, __tb_gg_main_pipe_1_gpu_gb, 0);
  static const size_t gg_main_pipe_1_gpu_gb_blocks = GG_MIN(blocks.x, ggc_get_nSM() * gg_main_pipe_1_gpu_gb_residency);
  if(!gg_main_pipe_1_gpu_gb_barrier_inited) { gg_main_pipe_1_gpu_gb_barrier.Setup(gg_main_pipe_1_gpu_gb_blocks); gg_main_pipe_1_gpu_gb_barrier_inited = true;};
  // FP: "2 -> 3;
  if (false)
  {
    gg_main_pipe_1(p2,p0,rp,iter,gg,hg,MAX_ITERATIONS,pipe,blocks,threads);
  }
  else
  {
    int* cl_iter;
    check_cuda(cudaMalloc(&cl_iter, sizeof(int) * 1));
    check_cuda(cudaMemcpy(cl_iter, &iter, sizeof(int) * 1, cudaMemcpyHostToDevice));

    // gg_main_pipe_1_gpu<<<1,1>>>(p2,p0,rp,iter,gg,hg,MAX_ITERATIONS,pipe,blocks,threads,cl_iter);
    gg_main_pipe_1_gpu_gb<<<gg_main_pipe_1_gpu_gb_blocks, __tb_gg_main_pipe_1_gpu_gb>>>(p2,p0,rp,iter,gg,hg,MAX_ITERATIONS,pipe,cl_iter, gg_main_pipe_1_gpu_gb_barrier);
    check_cuda(cudaMemcpy(&iter, cl_iter, sizeof(int) * 1, cudaMemcpyDeviceToHost));
    check_cuda(cudaFree(cl_iter));
  }
  // FP: "5 -> 6;
}
void gg_main(CSRGraphTex& hg, CSRGraphTex& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  static GlobalBarrierLifetime remove_dups_barrier;
  static bool remove_dups_barrier_inited;
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  Shared<float> p[3] = {Shared<float> (hg.nnodes), Shared<float> (hg.nnodes), Shared<float>(hg.nnodes)};
  // FP: "2 -> 3;
  Shared<float> r (hg.nnodes);
  // FP: "3 -> 4;
  Shared<int> marks (hg.nnodes);
  // FP: "4 -> 5;
  static const size_t remove_dups_residency = maximum_residency(remove_dups, __tb_remove_dups, 0);
  static const size_t remove_dups_blocks = GG_MIN(blocks.x, ggc_get_nSM() * remove_dups_residency);
  if(!remove_dups_barrier_inited) { remove_dups_barrier.Setup(remove_dups_blocks); remove_dups_barrier_inited = true;};
  // FP: "5 -> 6;
  int curr = 0;
  int iter = 0;
  float l1 = 0;
  // FP: "6 -> 7;
  r.zero_gpu();
  // FP: "7 -> 8;
  init_1 <<<blocks, threads>>>(gg, p[0].gpu_wr_ptr(), r.gpu_wr_ptr());
  // FP: "8 -> 9;
  gfloat_p p0 =p[0].gpu_wr_ptr();
  gfloat_p p2 =p[2].gpu_wr_ptr();
  gfloat_p rp =r.gpu_wr_ptr();
  // FP: "9 -> 10;
  pipe = PipeContextT<Worklist2>(hg.nedges);
  gg_main_pipe_1_wrapper(p2,p0,rp,iter,gg,hg,MAX_ITERATIONS,pipe,blocks,threads);
  pipe.free();
  // FP: "10 -> 11;
  printf("PR took %d iterations\n", iter);
  P_CURR = p[0].cpu_rd_ptr();
  P_NEXT = p[0].cpu_rd_ptr();
  // FP: "19 -> 20;
}