/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=texture";
#include "kernels/reduce.cuh"
typedef int edge_data_type;
typedef int node_data_type;
float* P_CURR ;
float* P_NEXT ;
extern const float ALPHA = 0.85;
extern const float EPSILON = 0.01;
extern int MAX_ITERATIONS ;
__global__ void init_1(CSRGraphTex graph, float * p_curr, float * residual)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    float update;
    index_type edge_end;
    p_curr[node] = 1.0 - ALPHA;
    update = 1.0/graph.getOutDegree(node);
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      dst = graph.getAbsDestination(edge);
      atomicAdd(residual + dst, update);
    }
  }
  // FP: "10 -> 11;
}
__global__ void init_2(CSRGraphTex graph, float * residual)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    residual[node] *= (1.0 - ALPHA) * ALPHA;
  }
  // FP: "4 -> 5;
}
__global__ void pagerank_main(CSRGraphTex graph, float * p_curr, float * residual, float * p_diff)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    int sdeg;
    float update;
    index_type edge_end;
    float res =atomicExch(residual + node, 0);
    p_curr[node] += res;
    p_diff[node] = res;
    sdeg = graph.getOutDegree(node);
    update = res * ALPHA / sdeg;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      float prev;
      dst = graph.getAbsDestination(edge);
      prev = atomicAdd(residual + dst, update);
    }
  }
  // FP: "12 -> 13;
}
void gg_main(CSRGraphTex& hg, CSRGraphTex& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  // FP: "1 -> 2;
  Shared<float> p[3] = {Shared<float> (hg.nnodes), Shared<float> (hg.nnodes), Shared<float>(hg.nnodes)};
  // FP: "2 -> 3;
  Shared<float> r (hg.nnodes);
  // FP: "3 -> 4;
  Shared<int> marks (hg.nnodes);
  // FP: "4 -> 5;
  int curr = 0;
  int iter = 0;
  float l1 = 0;
  // FP: "5 -> 6;
  r.zero_gpu();
  // FP: "6 -> 7;
  init_1 <<<blocks, threads>>>(gg, p[0].gpu_wr_ptr(), r.gpu_wr_ptr());
  // FP: "7 -> 8;
  init_2 <<<blocks, threads>>>(gg, r.gpu_wr_ptr());
  // FP: "8 -> 9;
  while (true)
  {
    pagerank_main <<<blocks, threads>>>(gg, p[0].gpu_wr_ptr(), r.gpu_wr_ptr(), p[2].gpu_wr_ptr());
    iter++;
    mgpu::Reduce(p[2].gpu_rd_ptr(), hg.nnodes, (float )0.0, mgpu::plus<float>(), (float*)0, &l1, *mgc);
    printf("%d: %f %d\n", iter, l1, l1 > EPSILON);
    p[2].zero_gpu();
    if (!(l1 > EPSILON) || (iter >= MAX_ITERATIONS))
    {
      break;
    }
  }
  // FP: "18 -> 19;
  printf("PR took %d iterations\n", iter);
  P_CURR = p[0].cpu_rd_ptr();
  P_NEXT = p[0].cpu_rd_ptr();
  // FP: "19 -> 20;
}