/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
#include "kernels/reduce.cuh"
typedef int edge_data_type;
typedef int node_data_type;
float* P_CURR ;
float* P_NEXT ;
extern const float ALPHA = 0.85;
extern const float EPSILON = 0.01;
extern int MAX_ITERATIONS ;
__global__ void init(CSRGraph graph, float * p_curr, float * p_next)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    p_curr[node] = 1.0-ALPHA;
    p_next[node] = 0.0;
  }
  // FP: "4 -> 5;
}
__global__ void pagerank_push(CSRGraph graph, float * p_curr, float * p_next)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    int sdeg;
    float update;
    index_type edge_end;
    sdeg = graph.getOutDegree(node);
    update = p_curr[node] / sdeg;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      dst = graph.getAbsDestination(edge);
      atomicAdd(p_next + dst, update);
    }
  }
  // FP: "10 -> 11;
}
__global__ void pagerank_compute(CSRGraph graph, float * p_curr, float * p_next, float * diff)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    p_next[node] = ALPHA * p_next[node] + (1.0 - ALPHA) / graph.nnodes;
    diff[node] = fabs(p_next[node] - p_curr[node]);
    p_curr[node] = 0.0;
  }
  // FP: "4 -> 5;
}
void gg_main(CSRGraph& hg, CSRGraph& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  // FP: "1 -> 2;
  Shared<float> p[3] = {Shared<float> (hg.nnodes), Shared<float> (hg.nnodes), Shared<float>(hg.nnodes)};
  // FP: "2 -> 3;
  int curr = 0;
  float l1 = 0.0;
  int iter = 0;
  // FP: "3 -> 4;
  init <<<blocks, threads>>>(gg, p[0].gpu_wr_ptr(), p[1].gpu_wr_ptr());
  // FP: "4 -> 5;
  do
  {
    pagerank_push <<<blocks, threads>>>(gg, p[curr].gpu_wr_ptr(), p[curr ^ 1].gpu_wr_ptr());
    pagerank_compute <<<blocks, threads>>>(gg, p[curr].gpu_wr_ptr(), p[curr ^ 1].gpu_wr_ptr(), p[2].gpu_wr_ptr(true));
    l1 = 0.0;
    iter++;
    curr ^= 1;
    mgpu::Reduce(p[2].gpu_rd_ptr(), hg.nnodes, (float )0.0, mgpu::plus<float>(), (float*)0, &l1, *mgc);
    printf("iteration: %d l1 = %f\n", iter, l1);
  }
  while (l1 > EPSILON && iter < MAX_ITERATIONS);
  // FP: "11 -> 12;
  printf("l1 = %f after %d iterations\n", l1, iter);
  P_CURR = p[curr].cpu_rd_ptr();
  P_NEXT = p[curr ^ 1].cpu_rd_ptr();
  // FP: "12 -> 13;
}