/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=texture";
typedef int edge_data_type;
typedef int node_data_type;
extern const node_data_type INF = INT_MAX;
__global__ void sssp_init(CSRGraphTex graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = (node == src) ? 0 : INF ;
  }
  // FP: "4 -> 5;
}
__global__ void sssp_kernel(CSRGraphTex graph, WorklistT in_wl, WorklistT out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    if(graph.node_data[node] == INF) continue;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst = graph.getAbsDestination(edge);
      edge_data_type wt = graph.getAbsWeight(edge);
      if (graph.node_data[dst] > graph.node_data[node] + wt)
      {
        atomicMin(graph.node_data + dst, graph.node_data[node] + wt);
        (out_wl).push(dst, graph.node_data[node] + wt);
      }
    }
  }
  // FP: "13 -> 14;
}
void gg_main(CSRGraphTex& hg, CSRGraphTex& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  PipeContextT<WorklistT> wl;
  // FP: "1 -> 2;
  sssp_init <<<blocks, threads>>>(gg, 0);
  // FP: "2 -> 3;
  int i = 0;
  // FP: "3 -> 4;
  wl = PipeContextT<WorklistT>(gg.nedges*2);
  wl.in_wl().wl[0] = 0;
  wl.in_wl().update_gpu(1);
  while (wl.in_wl().nitems())
  {
    wl.out_wl().will_write();
    sssp_kernel <<<blocks, threads>>>(gg, wl.in_wl(), wl.out_wl());
    wl.in_wl().swap_slots();
    wl.advance2();
    wl.in_wl().sort_prio();
    i++;
  }
  wl.free();
  // FP: "6 -> 7;
  printf("iterations: %d\n", i);
  // FP: "7 -> 8;
}