/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=8 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=True $ ignore_nested_errors=False $ np=True $ write_props=None $ quiet_cgen=True $ retry_backoff=True $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=texture";
#include <kernels/segmentedsort.cuh>
void debug_output(CSRGraphTy &g, unsigned int *valid_edges);;
static const int __tb_preprocess = TB_SIZE;
static const int __tb_count_triangles = TB_SIZE;
__global__ void preprocess(CSRGraphTex graph, unsigned int * valid_edges)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_preprocess;
  index_type node_end;
  index_type node_rup;
  // FP: "1 -> 2;
  const int _NP_CROSSOVER_WP = 32;
  const int _NP_CROSSOVER_TB = __kernel_tb_size;
  // FP: "2 -> 3;
  const int BLKSIZE = __kernel_tb_size;
  const int ITSIZE = BLKSIZE * 8;
  // FP: "3 -> 4;

  typedef cub::BlockScan<index_type, BLKSIZE> BlockScan;
  typedef union np_shared<BlockScan::TempStorage, index_type, struct empty_np, struct empty_np, struct fg_np<ITSIZE> > npsTy;

  // FP: "4 -> 5;
  __shared__ npsTy nps ;
  // FP: "5 -> 6;
  const index_type last  = graph.nnodes;
  // FP: "6 -> 7;
  node_end = (graph).nnodes;
  node_rup = ((0) + roundup((((graph).nnodes) - (0)), (blockDim.x)));
  for (index_type node = 0 + tid; node < node_rup; node += nthreads)
  {
    bool pop;
    int degree;
    index_type _np_mps;
    index_type _np_mps_total;
    // FP: "7 -> 8;
    // FP: "8 -> 9;
    pop = graph.valid_node(node);;
    // FP: "9 -> 10;
    if (pop)
    {
      degree = graph.getOutDegree(node);
    }
    // FP: "12 -> 13;
    struct NPInspector1 _np = {0,0,0,0,0,0};
    // FP: "13 -> 14;
    __shared__ struct { index_type node; int degree; } _np_closure [TB_SIZE];
    // FP: "14 -> 15;
    _np_closure[threadIdx.x].node = node;
    _np_closure[threadIdx.x].degree = degree;
    // FP: "15 -> 16;
    if (pop)
    {
      _np.size = (graph).getOutDegree(node);
      _np.start = (graph).getFirstEdge(node);
    }
    // FP: "18 -> 19;
    // FP: "19 -> 20;
    _np_mps = _np.size;
    // FP: "20 -> 21;
    BlockScan(nps.temp_storage).ExclusiveSum(_np_mps, _np_mps, _np_mps_total);
    // FP: "21 -> 22;
    if (threadIdx.x == 0)
    {
    }
    // FP: "23 -> 24;
    __syncthreads();
    // FP: "24 -> 25;
    _np.total = _np_mps_total;
    _np.offset = _np_mps;
    // FP: "25 -> 26;
    while (_np.work())
    {
      // FP: "26 -> 27;
      int _np_i =0;
      // FP: "27 -> 28;
      _np.inspect2(nps.fg.itvalue, nps.fg.src, ITSIZE, threadIdx.x);
      // FP: "28 -> 29;
      __syncthreads();
      // FP: "29 -> 30;

      // FP: "30 -> 31;
      for (_np_i = threadIdx.x; _np_i < ITSIZE && _np.valid(_np_i); _np_i += BLKSIZE)
      {
        index_type edge;
        assert(nps.fg.src[_np_i] < __kernel_tb_size);
        node = _np_closure[nps.fg.src[_np_i]].node;
        degree = _np_closure[nps.fg.src[_np_i]].degree;
        edge= nps.fg.itvalue[_np_i];
        {
          index_type dst = graph.getAbsDestination(edge);
          int dst_degree = graph.getOutDegree(dst);
          if ((dst_degree > degree) || (dst_degree == degree && dst > node))
          {
            graph.edge_data[edge] = dst;
            atomicAdd(valid_edges + node, 1);
          }
          else
          {
            graph.edge_data[edge] = graph.nnodes;
          }
        }
      }
      // FP: "40 -> 41;
      _np.execute_round_done(ITSIZE);
      // FP: "41 -> 42;
      __syncthreads();
    }
    // FP: "43 -> 44;
    assert(threadIdx.x < __kernel_tb_size);
    node = _np_closure[threadIdx.x].node;
    degree = _np_closure[threadIdx.x].degree;
  }
  // FP: "45 -> 46;
}
__device__ unsigned int intersect(CSRGraphTex graph, index_type u, index_type v, unsigned int * valid_edges)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  // FP: "1 -> 2;
  index_type u_start = graph.getFirstEdge(u);
  index_type u_end = u_start + valid_edges[u];
  index_type v_start = graph.getFirstEdge(v);
  index_type v_end = v_start + valid_edges[v];
  int count = 0;
  index_type u_it = u_start;
  index_type v_it = v_start;
  index_type a ;
  index_type b ;
  // FP: "2 -> 3;
  while (u_it < u_end && v_it < v_end)
  {
    a = graph.getAbsDestination(u_it);
    b = graph.getAbsDestination(v_it);
    int d = a - b;
    if (d <= 0)
    {
      u_it++;
    }
    if (d >= 0)
    {
      v_it++;
    }
    if (d == 0)
    {
      count++;
    }
  }
  // FP: "16 -> 17;
  return count;
  // FP: "17 -> 18;
}
__global__ void count_triangles(CSRGraphTex graph, unsigned int * valid_edges, int * count)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = __tb_count_triangles;
  index_type v_end;
  index_type v_rup;
  // FP: "1 -> 2;
  const int _NP_CROSSOVER_WP = 32;
  const int _NP_CROSSOVER_TB = __kernel_tb_size;
  // FP: "2 -> 3;
  const int BLKSIZE = __kernel_tb_size;
  const int ITSIZE = BLKSIZE * 8;
  // FP: "3 -> 4;

  typedef cub::BlockScan<index_type, BLKSIZE> BlockScan;
  typedef union np_shared<BlockScan::TempStorage, index_type, struct empty_np, struct empty_np, struct fg_np<ITSIZE> > npsTy;

  // FP: "4 -> 5;
  __shared__ npsTy nps ;
  // FP: "5 -> 6;
  int lcount =0;
  // FP: "6 -> 7;
  v_end = (graph).nnodes;
  v_rup = ((0) + roundup((((graph).nnodes) - (0)), (blockDim.x)));
  for (index_type v = 0 + tid; v < v_rup; v += nthreads)
  {
    bool pop;
    int d_v;
    index_type _np_mps;
    index_type _np_mps_total;
    // FP: "7 -> 8;
    // FP: "8 -> 9;
    pop = graph.valid_node(v);;
    // FP: "9 -> 10;
    struct NPInspector1 _np = {0,0,0,0,0,0};
    // FP: "10 -> 11;
    __shared__ struct { index_type v; } _np_closure [TB_SIZE];
    // FP: "11 -> 12;
    _np_closure[threadIdx.x].v = v;
    // FP: "12 -> 13;
    if (pop)
    {
      _np.size = ((graph).getFirstEdge(v)+ valid_edges[v]) - ((graph).getFirstEdge(v));
      _np.start = (graph).getFirstEdge(v);
    }
    // FP: "15 -> 16;
    // FP: "16 -> 17;
    _np_mps = _np.size;
    // FP: "17 -> 18;
    BlockScan(nps.temp_storage).ExclusiveSum(_np_mps, _np_mps, _np_mps_total);
    // FP: "18 -> 19;
    if (threadIdx.x == 0)
    {
    }
    // FP: "20 -> 21;
    __syncthreads();
    // FP: "21 -> 22;
    _np.total = _np_mps_total;
    _np.offset = _np_mps;
    // FP: "22 -> 23;
    while (_np.work())
    {
      // FP: "23 -> 24;
      int _np_i =0;
      // FP: "24 -> 25;
      _np.inspect2(nps.fg.itvalue, nps.fg.src, ITSIZE, threadIdx.x);
      // FP: "25 -> 26;
      __syncthreads();
      // FP: "26 -> 27;

      // FP: "27 -> 28;
      for (_np_i = threadIdx.x; _np_i < ITSIZE && _np.valid(_np_i); _np_i += BLKSIZE)
      {
        index_type edge;
        assert(nps.fg.src[_np_i] < __kernel_tb_size);
        v = _np_closure[nps.fg.src[_np_i]].v;
        edge= nps.fg.itvalue[_np_i];
        {
          index_type u = graph.getAbsDestination(edge);
          index_type d_u = graph.getOutDegree(u);
          int xcount = 0;
          xcount = intersect(graph, u, v, valid_edges);
          if (xcount)
          {
            atomicAdd(count, xcount);
          }
        }
      }
      // FP: "37 -> 38;
      _np.execute_round_done(ITSIZE);
      // FP: "38 -> 39;
      __syncthreads();
    }
    // FP: "40 -> 41;
    assert(threadIdx.x < __kernel_tb_size);
    v = _np_closure[threadIdx.x].v;
  }
  // FP: "42 -> 43;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  // FP: "1 -> 2;
  Shared<int> count (1);
  Shared<unsigned int> valid_edges (hg.nnodes);
  // FP: "2 -> 3;
  count.zero_gpu();
  // FP: "3 -> 4;
  valid_edges.zero_gpu();
  // FP: "4 -> 5;
  preprocess <<<blocks, __tb_preprocess>>>(gg, valid_edges.gpu_wr_ptr());
  // FP: "5 -> 6;
  mgpu::SegSortPairsFromIndices(gg.edge_data, gg.edge_dst, gg.nedges, (const int *) gg.row_start + 1, gg.nnodes - 1, *mgc);;
  // FP: "6 -> 7;
  count_triangles <<<blocks, __tb_count_triangles>>>(gg, valid_edges.gpu_rd_ptr(), count.gpu_wr_ptr());
  // FP: "7 -> 8;
  printf("triangles: %d\n", *count.cpu_rd_ptr());
  // FP: "8 -> 9;
}