/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
typedef int edge_data_type;
typedef int node_data_type;
#define RESIDENCY 6
#define TB_SIZE_CX 128
#define WORKLISTSIZE 14336*2
extern const node_data_type INF = INT_MAX;
static const int __tb_contract = TB_SIZE;
static const int __tb_filter = TB_SIZE;
static const int __tb_contract_expand = TB_SIZE_CX;
static const int __tb_expand = TB_SIZE;
__global__ void kernel(CSRGraph graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = INF ;
  }
  // FP: "4 -> 5;
}
__global__ void __launch_bounds__(TB_SIZE_CX, RESIDENCY) contract_expand(CSRGraph graph, int LEVEL, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  index_type wlnode_rup;
  // FP: "1 -> 2;
  const int WHTSIZE = 64;
  __shared__ volatile int whash[__kernel_tb_size/32][WHTSIZE] ;
  // FP: "2 -> 3;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  wlnode_rup = ((0) + roundup(((*((volatile index_type *) (in_wl).dindex)) - (0)), (blockDim.x)));
  for (index_type wlnode = 0 + tid; wlnode < wlnode_rup; wlnode += nthreads)
  {
    int node;
    bool pop;
    index_type edge_end;
    // FP: "3 -> 4;
    // FP: "4 -> 5;
    pop = (in_wl).pop_id(wlnode, node);
    // FP: "5 -> 6;
    if (pop)
    {
      pop = false;
      if (graph.node_data[node] == INF)
      {
        graph.node_data[node] = LEVEL;
        pop = true;
      }
    }
    // FP: "11 -> 12;
    if (pop)
    {
      const int warpid = threadIdx.x / 32;
      const int htentry = node & (WHTSIZE - 1);
      whash[warpid][htentry] = node;
      if (whash[warpid][htentry] == node)
      {
        whash[warpid][htentry] = threadIdx.x;
        pop = whash[warpid][htentry] == threadIdx.x;
      }
    }
    // FP: "18 -> 19;
    __syncthreads();
    // FP: "19 -> 20;
    if (!pop)
    {
      continue;
    }
    // FP: "22 -> 23;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      dst = graph.getAbsDestination(edge);
      (out_wl).push(dst);
    }
  }
  // FP: "28 -> 29;
}
__global__ void __launch_bounds__(TB_SIZE, 8) expand(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      dst = graph.getAbsDestination(edge);
      (out_wl).push(dst);
    }
  }
  // FP: "10 -> 11;
}
__global__ void __launch_bounds__(TB_SIZE, 8) contract(CSRGraph graph, int LEVEL, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop)
    {
      if (graph.node_data[node] == INF)
      {
        graph.node_data[node] = LEVEL;
        (out_wl).push(node);
      }
    }
  }
  // FP: "11 -> 12;
}
__global__ void __launch_bounds__(TB_SIZE, 8) filter(ApproxBitsetByte visited, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop)
    {
      if (!visited.is_set(node))
      {
        visited.set(node);
        (out_wl).push(node);
      }
    }
  }
  // FP: "11 -> 12;
}
void gg_main(CSRGraph& hg, CSRGraph& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  ApproxBitsetByte visited (hg.nnodes);
  // FP: "2 -> 3;
  kernel <<<blocks, threads>>>(gg, 0);
  // FP: "3 -> 4;
  int LEVEL = 0;
  // FP: "4 -> 5;
  pipe = PipeContextT<Worklist2>(gg.nedges > 65536 ? gg.nedges : 65536);
  pipe.in_wl().wl[0] = 0;
  pipe.in_wl().update_gpu(1);
  {
    while (pipe.in_wl().nitems())
    {
      while (pipe.in_wl().nitems())
      {
        pipe.out_wl().will_write();
        contract_expand <<<blocks, __tb_contract_expand>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        LEVEL++;
        if (pipe.in_wl().nitems() > WORKLISTSIZE)
        {
          break;
        }
      }
      while (pipe.in_wl().nitems())
      {
        pipe.out_wl().will_write();
        contract <<<blocks, __tb_contract>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        pipe.out_wl().will_write();
        expand <<<blocks, __tb_expand>>>(gg, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        pipe.out_wl().will_write();
        filter <<<blocks, __tb_filter>>>(visited, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        LEVEL++;
      }
    }
  }
  pipe.free();
  // FP: "19 -> 20;
}