/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
typedef int edge_data_type;
typedef int node_data_type;
extern const node_data_type INF = INT_MAX;
__global__ void bfs_init(CSRGraph graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = (node == src) ? 0 : INF ;
  }
  // FP: "4 -> 5;
}
__global__ void bfs_topo(CSRGraph graph, int LEVEL, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    index_type edge_end;
    if(graph.node_data[node] != (LEVEL - 1)) continue;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst = graph.getAbsDestination(edge);
      if (graph.node_data[dst] == INF)
      {
        graph.node_data[dst] = LEVEL;
        ret_val.return_(true);
        continue;
      }
    }
  }
}
void gg_main(CSRGraph& hg, CSRGraph& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  bool loopc = false;
  // FP: "1 -> 2;
  bfs_init <<<blocks, threads>>>(gg, 0);
  // FP: "2 -> 3;
  int LEVEL = 1;
  // FP: "3 -> 4;
  do
  {
    Shared<int> retval = Shared<int>(1);
    Any _rv;
    *(retval.cpu_wr_ptr()) = 0;
    _rv.rv = retval.gpu_wr_ptr();
    bfs_topo <<<blocks, threads>>>(gg, LEVEL, _rv);
    loopc = *(retval.cpu_rd_ptr()) > 0;
    LEVEL++;
  }
  while (loopc);
  // FP: "6 -> 7;
}