/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=texture";
typedef int edge_data_type;
typedef int node_data_type;
extern const node_data_type INF = INT_MAX;
static const int __tb_filter = TB_SIZE;
static const int __tb_contract = TB_SIZE;
static const int __tb_expand = TB_SIZE;
__global__ void kernel(CSRGraphTex graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = INF ;
  }
  // FP: "4 -> 5;
}
__global__ void __launch_bounds__(TB_SIZE, 8) expand(CSRGraphTex graph, WorklistT in_wl, WorklistT out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      dst = graph.getAbsDestination(edge);
      (out_wl).push(dst);
    }
  }
  // FP: "10 -> 11;
}
__global__ void __launch_bounds__(TB_SIZE, 8) contract(CSRGraphTex graph, int LEVEL, WorklistT in_wl, WorklistT out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop)
    {
      if (graph.node_data[node] == INF)
      {
        graph.node_data[node] = LEVEL;
        (out_wl).push(node);
      }
    }
  }
  // FP: "11 -> 12;
}
__global__ void __launch_bounds__(TB_SIZE, 8) filter(ApproxBitsetByte visited, WorklistT in_wl, WorklistT out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop)
    {
      if (!visited.is_set(node))
      {
        visited.set(node);
        (out_wl).push(node);
      }
    }
  }
  // FP: "11 -> 12;
}
void gg_main(CSRGraphTex& hg, CSRGraphTex& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  static GlobalBarrierLifetime expand_barrier;
  static bool expand_barrier_inited;
  PipeContextT<WorklistT> pipe;
  // FP: "1 -> 2;
  static const size_t expand_residency = maximum_residency(expand, __tb_expand, 0);
  static const size_t expand_blocks = GG_MIN(blocks.x, ggc_get_nSM() * expand_residency);
  if(!expand_barrier_inited) { expand_barrier.Setup(expand_blocks); expand_barrier_inited = true;};
  // FP: "2 -> 3;
  kernel <<<blocks, threads>>>(gg, 0);
  // FP: "3 -> 4;
  int LEVEL = 0;
  // FP: "4 -> 5;
  ApproxBitsetByte visited (hg.nnodes);
  // FP: "5 -> 6;
  pipe = PipeContextT<WorklistT>(gg.nedges > 65536 ? gg.nedges : 65536);
  pipe.in_wl().wl[0] = 0;
  pipe.in_wl().update_gpu(1);
  {
    while (pipe.in_wl().nitems())
    {
      pipe.out_wl().will_write();
      contract <<<blocks, __tb_contract>>>(gg, LEVEL, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      pipe.out_wl().will_write();
      expand <<<blocks, __tb_expand>>>(gg, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      pipe.out_wl().will_write();
      filter <<<blocks, __tb_filter>>>(visited, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      LEVEL++;
    }
  }
  pipe.free();
  // FP: "11 -> 12;
}