/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=texture";
extern unsigned long DISCOUNT_TIME_NS;
__global__ void init(CSRGraphTex graph)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = node;
  }
  // FP: "4 -> 5;
}
__global__ void prep_edge_src(CSRGraphTex graph, index_type * edge_src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    bool pop;
    index_type edge_end;
    pop = graph.valid_node(node);;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      edge_src[edge] = node;
    }
  }
  // FP: "8 -> 9;
}
__global__ void hook_init(CSRGraphTex graph, index_type * edge_src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  int edge_end;
  // FP: "1 -> 2;
  edge_end = graph.nedges;
  for (int edge = 0 + tid; edge < edge_end; edge += nthreads)
  {
    index_type x = edge_src[edge];
    index_type y = graph.getAbsDestination(edge);
    index_type mx = x > y ? x : y;
    index_type mn = x > y ? y : x;
    graph.node_data[mx] = mn;
  }
  // FP: "6 -> 7;
}
__global__ void hook_high_to_low(CSRGraphTex graph, const __restrict__ index_type * edge_src, char * marks, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  int edge_end;
  // FP: "1 -> 2;
  edge_end = graph.nedges;
  for (int edge = 0 + tid; edge < edge_end; edge += nthreads)
  {
    if (!marks[edge])
    {
      index_type u = edge_src[edge];
      index_type v = graph.getAbsDestination(edge);
      node_data_type p_u = graph.node_data[u];
      node_data_type p_v = graph.node_data[v];
      index_type mx = p_u > p_v ? p_u : p_v;
      index_type mn = p_u > p_v ? p_v : p_u;
      if (mx == mn)
      {
        marks[edge] = 1;
      }
      else
      {
        graph.node_data[mn] = mx;
        ret_val.return_(true);
        continue;
      }
    }
  }
}
__global__ void hook_low_to_high(CSRGraphTex graph, index_type * edge_src, char * marks, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  int edge_end;
  // FP: "1 -> 2;
  edge_end = graph.nedges;
  for (int edge = 0 + tid; edge < edge_end; edge += nthreads)
  {
    if (!marks[edge])
    {
      index_type u = edge_src[edge];
      index_type v = graph.getAbsDestination(edge);
      node_data_type p_u = graph.node_data[u];
      node_data_type p_v = graph.node_data[v];
      index_type mx = p_u > p_v ? p_u : p_v;
      index_type mn = p_u > p_v ? p_v : p_u;
      if (mx == mn)
      {
        marks[edge] = 1;
      }
      else
      {
        graph.node_data[mx] = mn;
        ret_val.return_(true);
        continue;
      }
    }
  }
}
__global__ void p_jump(CSRGraphTex graph, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    node_data_type p_u = graph.node_data[node];
    node_data_type p_v = graph.node_data[p_u];
    if (p_u != p_v)
    {
      graph.node_data[node] = p_v;
      ret_val.return_(true);
      continue;
    }
  }
}
__global__ void identify_roots(CSRGraphTex graph, WorklistT in_wl, WorklistT out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (graph.node_data[node] == node)
    {
      (out_wl).push(node);
    }
  }
  // FP: "6 -> 7;
}
__global__ void p_jump_roots(CSRGraphTex graph, WorklistT in_wl, WorklistT out_wl, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    pop = (in_wl).pop_id(wlnode, node);
    node_data_type p_u = graph.node_data[node];
    node_data_type p_v = graph.node_data[p_u];
    if (p_u != p_v)
    {
      graph.node_data[node] = p_v;
      ret_val.return_(true);
      continue;
    }
  }
}
__global__ void p_jump_leaves(CSRGraphTex graph)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    node_data_type p_u = graph.node_data[node];
    node_data_type p_v = graph.node_data[p_u];
    if (p_u != p_v)
    {
      graph.node_data[node] = p_v;
    }
  }
  // FP: "7 -> 8;
}
__global__ void count_components(CSRGraphTex graph, int * count)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (node == graph.node_data[node])
    {
      atomicAdd(count, 1);
    }
  }
  // FP: "6 -> 7;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  int edge_blocks;
  int node_blocks;
  cudaEvent_t start;
  cudaEvent_t stop;
  bool loopc = false;
  PipeContextT<WorklistT> pipe;
  // FP: "1 -> 2;
  int it_hk = 1;
  // FP: "2 -> 3;
  Shared<index_type> edge_src (gg.nedges);
  // FP: "3 -> 4;
  Shared<char> edge_marks (gg.nedges);
  // FP: "4 -> 5;
  bool flag = false;
  // FP: "5 -> 6;
  // FP: "6 -> 7;
  edge_blocks = hg.nedges / TB_SIZE + 1;
  node_blocks = hg.nnodes / TB_SIZE + 1;
  // FP: "7 -> 8;
  edge_marks.zero_gpu();
  // FP: "8 -> 9;
  // FP: "9 -> 10;
  check_cuda(cudaEventCreate(&start));
  check_cuda(cudaEventCreate(&stop));
  // FP: "10 -> 11;
  check_cuda(cudaEventRecord(start));
  // FP: "11 -> 12;
  prep_edge_src <<<node_blocks, threads>>>(gg, edge_src.gpu_wr_ptr());
  // FP: "12 -> 13;
  check_cuda(cudaEventRecord(stop));
  // FP: "13 -> 14;
  init <<<node_blocks, threads>>>(gg);
  // FP: "14 -> 15;
  hook_init <<<edge_blocks, threads>>>(gg, edge_src.gpu_rd_ptr());
  // FP: "15 -> 16;
  do
  {
    Shared<int> retval = Shared<int>(1);
    Any _rv;
    *(retval.cpu_wr_ptr()) = 0;
    _rv.rv = retval.gpu_wr_ptr();
    p_jump <<<node_blocks, threads>>>(gg, _rv);
    loopc = *(retval.cpu_rd_ptr()) > 0;
    // FP: "16 -> 16;
  }
  while (loopc);
  // FP: "17 -> 18;
  pipe = PipeContextT<WorklistT>(gg.nnodes);
  {
    {
      do
      {
        bool loopc = false;
        pipe.out_wl().will_write();
        identify_roots <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        if (it_hk != 0)
        {
          Shared<int> retval = Shared<int>(1);
          Any _rv;
          *(retval.cpu_wr_ptr()) = 0;
          _rv.rv = retval.gpu_wr_ptr();
          hook_low_to_high <<<edge_blocks, threads>>>(gg, edge_src.gpu_rd_ptr(), edge_marks.gpu_wr_ptr(), _rv);
          flag = *(retval.cpu_rd_ptr());
          it_hk = (it_hk + 1) % 4;
        }
        else
        {
          Shared<int> retval = Shared<int>(1);
          Any _rv;
          *(retval.cpu_wr_ptr()) = 0;
          _rv.rv = retval.gpu_wr_ptr();
          hook_high_to_low <<<edge_blocks, threads>>>(gg, edge_src.gpu_rd_ptr(), edge_marks.gpu_wr_ptr(), _rv);
          flag = *(retval.cpu_rd_ptr());
        }
        if (!flag)
        {
          break;
        }
        do
        {
          Shared<int> retval = Shared<int>(1);
          Any _rv;
          *(retval.cpu_wr_ptr()) = 0;
          _rv.rv = retval.gpu_wr_ptr();
          pipe.out_wl().will_write();
          p_jump_roots <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl(), _rv);
          loopc = *(retval.cpu_rd_ptr()) > 0;
        }
        while (loopc);
        p_jump_leaves <<<node_blocks, threads>>>(gg);
      }
      while (flag);
      // FP: "36 -> 37;
    }
  }
  pipe.free();
  // FP: "18 -> 19;
  printf("iterations: %d\n", it_hk);
  // FP: "37 -> 38;
  Shared<int> count (1);
  // FP: "38 -> 39;
  *(count.cpu_wr_ptr()) = 0;
  // FP: "39 -> 40;
  count_components <<<blocks, threads>>>(gg, count.gpu_wr_ptr());
  // FP: "40 -> 41;
  printf("components: %d\n", *(count.cpu_rd_ptr()));
  // FP: "41 -> 42;
  float ms =0;
  // FP: "42 -> 43;
  check_cuda(cudaEventElapsedTime(&ms, start, stop));
  DISCOUNT_TIME_NS = (int) (ms * 1000000);
  printf("prep_edge_src: %llu ns\n", DISCOUNT_TIME_NS);
  // FP: "43 -> 44;
}