/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=texture";
extern int N_MIS_NODES;
extern int* MIS_NODES;
#define UNMARKED 0
#define MARKED 1
#define NON_INDEPENDENT 2
#define NON_MAXIMAL 3
__global__ void init_nodes(CSRGraphTex graph, WorklistT in_wl, WorklistT out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    pop = (in_wl).pop_id(wlnode, node);
    graph.node_data[node] = MARKED;
  }
  // FP: "6 -> 7;
}
__global__ void check_independence(CSRGraphTex graph, WorklistT in_wl, WorklistT out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst = graph.getAbsDestination(edge);
      if (dst != node && graph.node_data[dst] == MARKED)
      {
        graph.node_data[node] = NON_INDEPENDENT;
      }
    }
  }
  // FP: "11 -> 12;
}
__global__ void check_maximal(CSRGraphTex graph)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (graph.node_data[node] == UNMARKED)
    {
      index_type edge_end;
      bool marked_neighbour = false;
      edge_end = (graph).getFirstEdge((node) + 1);
      for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
      {
        index_type dst = graph.getAbsDestination(edge);
        if (dst != node && graph.node_data[dst] != UNMARKED)
        {
          marked_neighbour = true;
        }
      }
      if (!marked_neighbour)
      {
        graph.node_data[node] = NON_MAXIMAL;
      }
    }
  }
  // FP: "15 -> 16;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  PipeContextT<WorklistT> pipe;
  // FP: "1 -> 2;
  pipe = PipeContextT<WorklistT>(hg.nnodes);
  memcpy(pipe.in_wl().wl, MIS_NODES, (N_MIS_NODES) * sizeof((MIS_NODES)[0]));
  pipe.in_wl().update_gpu(N_MIS_NODES);
  {
    {
      pipe.out_wl().will_write();
      init_nodes <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
      // FP: "3 -> 4;
      pipe.out_wl().will_write();
      check_independence <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
      // FP: "4 -> 5;
      check_maximal <<<blocks, threads>>>(gg);
      // FP: "5 -> 6;
    }
  }
  pipe.free();
  // FP: "2 -> 3;
}