/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
#define ST_NP -1
#define ST_INACTIVE -2
#define ST_IND 2
#define SEED1 0x12345678LL
#define SEED2 0xabbdef12LL
#define SEED3 0xcafe1234LL
#define SEED4 0x09832516LL
__global__ void gen_prio_gpu(CSRGraph graph, unsigned int * prio, unsigned int x, unsigned int y, unsigned int z, unsigned int w)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  x ^= tid;
  y ^= tid;
  z ^= tid;
  w ^= tid;
  assert(!(x == 0 && y == 0 && z == 0 && w == 0));
  // FP: "2 -> 3;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    unsigned int t;
    t = x ^ (x << 11);
    x = y;
    y = z;
    z = w;
    w = w ^ (w >> 19) ^ t ^ (t >> 8);
    prio[node] = w;
    if (w == UINT_MAX)
    {
      prio[node] = UINT_MAX - 1;
    }
  }
  // FP: "9 -> 10;
}
__global__ void init(CSRGraph graph, int * c_array, int * cu_array)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    c_array[node] = ST_NP;
    cu_array[node] = ST_NP;
    graph.node_data[node] = 0;
  }
  // FP: "6 -> 7;
}
__global__ void mis1(CSRGraph graph, unsigned int * prio, int * c_array, unsigned int * min_array, Any ret_val)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (c_array[node] == ST_NP)
    {
      index_type edge_end;
      unsigned int minn = UINT_MAX;
      edge_end = (graph).getFirstEdge((node) + 1);
      for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
      {
        index_type dst =graph.getAbsDestination(edge);
        if (dst != node && c_array[dst] == ST_NP)
        {
          if (prio[dst] < minn)
          {
            minn = prio[dst];
          }
        }
      }
      min_array[node] = minn;
      ret_val.return_(true);
      continue;
    }
  }
}
__global__ void mis2(CSRGraph graph, unsigned int * prio, int * c_array, int * cu_array, unsigned int * min_array)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (prio[node] < min_array[node] && c_array[node] == ST_NP)
    {
      index_type edge_end;
      graph.node_data[node] = ST_IND;
      c_array[node] = ST_INACTIVE;
      edge_end = (graph).getFirstEdge((node) + 1);
      for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
      {
        index_type dst = graph.getAbsDestination(edge);
        if (c_array[dst] == ST_NP)
        {
          cu_array[dst] = ST_INACTIVE;
        }
      }
    }
  }
  // FP: "11 -> 12;
}
__global__ void mis3(CSRGraph graph, int * cu_array, int * c_array)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    if (cu_array[node] == ST_INACTIVE)
    {
      c_array[node] = cu_array[node];
    }
  }
  // FP: "6 -> 7;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  bool loopc = false;
  // FP: "1 -> 2;
  Shared<unsigned int> prio (hg.nnodes);
  int STEPS = 0;
  // FP: "2 -> 3;
  ggc::Timer t ("random");
  // FP: "3 -> 4;
  t.start();
  // FP: "4 -> 5;
  gen_prio_gpu <<<blocks, threads>>>(gg, prio.gpu_wr_ptr(), SEED1, SEED2, SEED3, SEED4);
  // FP: "5 -> 6;
  cudaDeviceSynchronize();
  // FP: "6 -> 7;
  t.stop();
  printf("Random number generation took %llu ns\n", t.duration());
  // FP: "7 -> 8;
  Shared<int> c_array (hg.nnodes);
  Shared<int> cu_array (hg.nnodes);
  Shared<unsigned int> min_array (hg.nnodes);
  // FP: "8 -> 9;
  int threads_1t  = 128;
  int blocks_1t  = (hg.nnodes + threads_1t - 1) / threads_1t;
  // FP: "9 -> 10;
  init <<<blocks_1t, threads>>>(gg, c_array.gpu_wr_ptr(), cu_array.gpu_wr_ptr());
  // FP: "10 -> 11;
  do
  {
    Shared<int> retval = Shared<int>(1);
    Any _rv;
    *(retval.cpu_wr_ptr()) = 0;
    _rv.rv = retval.gpu_wr_ptr();
    mis1 <<<blocks_1t, threads>>>(gg, prio.gpu_wr_ptr(), c_array.gpu_wr_ptr(), min_array.gpu_wr_ptr(), _rv);
    loopc = *(retval.cpu_rd_ptr()) > 0;
    mis2 <<<blocks_1t, threads>>>(gg, prio.gpu_wr_ptr(), c_array.gpu_wr_ptr(), cu_array.gpu_wr_ptr(), min_array.gpu_rd_ptr());
    mis3 <<<blocks_1t, threads>>>(gg, cu_array.gpu_rd_ptr(), c_array.gpu_wr_ptr());
    STEPS++;
  }
  while (loopc);
  // FP: "15 -> 16;
  printf("Total steps: %d\n", STEPS);
  // FP: "16 -> 17;
}