/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
#include <curand.h>
#define UNMARKED 0
#define MARKED 1
#define NON_INDEPENDENT 2
#define NON_MAXIMAL 3
#define SEED1 0x12345678LL
#define SEED2 0xabbdef12LL
#define SEED3 0xcafe1234LL
#define SEED4 0x09832516LL
__global__ void gen_prio_gpu(CSRGraph graph, unsigned int * prio, unsigned int x, unsigned int y, unsigned int z, unsigned int w)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  x ^= tid;
  y ^= tid;
  z ^= tid;
  w ^= tid;
  assert(!(x == 0 && y == 0 && z == 0 && w == 0));
  // FP: "2 -> 3;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    unsigned int t;
    t = x ^ (x << 11);
    x = y;
    y = z;
    z = w;
    w = w ^ (w >> 19) ^ t ^ (t >> 8);
    prio[node] = w;
  }
  // FP: "6 -> 7;
}
void gen_prio(CSRGraph graph, unsigned int * prio)
{
  curandGenerator_t gen;
  // FP: "1 -> 2;
  // FP: "2 -> 3;
  check_rv(curandCreateGenerator(&gen, CURAND_RNG_PSEUDO_MT19937), CURAND_STATUS_SUCCESS);
  check_rv(curandSetPseudoRandomGeneratorSeed(gen, SEED1), CURAND_STATUS_SUCCESS);
  check_rv(curandSetGeneratorOrdering (gen, CURAND_ORDERING_PSEUDO_BEST), CURAND_STATUS_SUCCESS);
  // FP: "3 -> 4;
}
__global__ void init_wl(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    (out_wl).push(node);
  }
  // FP: "4 -> 5;
}
__global__ void mark_nodes(CSRGraph graph, const unsigned int * __restrict__ prio, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    int max_prio = prio[node];
    int max_prio_node = node;
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst = graph.getAbsDestination(edge);
      if (dst != node && graph.node_data[dst] != NON_INDEPENDENT && prio[dst] >= max_prio)
      {
        if ((prio[dst] > max_prio) || dst > max_prio_node)
        {
          max_prio = prio[dst];
          max_prio_node = dst;
        }
      }
    }
    if (max_prio_node == node)
    {
      assert(graph.node_data[node] == UNMARKED);
      graph.node_data[node] = MARKED;
    }
  }
  // FP: "17 -> 18;
}
__global__ void drop_marked_nodes_and_nbors(CSRGraph graph, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    bool pop;
    int node;
    pop = (in_wl).pop_id(wlnode, node);
    bool drop = false;
    if (graph.node_data[node] == MARKED)
    {
      drop = true;
    }
    if (!drop)
    {
      index_type edge_end;
      edge_end = (graph).getFirstEdge((node) + 1);
      for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
      {
        index_type dst = graph.getAbsDestination(edge);
        if (graph.node_data[dst] == MARKED)
        {
          drop = true;
        }
      }
    }
    if (!drop)
    {
      (out_wl).push(node);
    }
    else
    {
      if (graph.node_data[node] == UNMARKED)
      {
        graph.node_data[node] = NON_INDEPENDENT;
      }
    }
  }
  // FP: "23 -> 24;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  Shared<unsigned int> prio (hg.nnodes);
  int STEPS = 0;
  // FP: "2 -> 3;
  ggc::Timer t ("random");
  // FP: "3 -> 4;
  t.start();
  // FP: "4 -> 5;
  gen_prio_gpu <<<blocks, threads>>>(gg, prio.gpu_wr_ptr(), SEED1, SEED2, SEED3, SEED4);
  // FP: "5 -> 6;
  cudaDeviceSynchronize();
  // FP: "6 -> 7;
  t.stop();
  printf("Random number generation took %llu ns\n", t.duration());
  // FP: "7 -> 8;
  pipe = PipeContextT<Worklist2>(gg.nnodes);
  {
    {
      pipe.out_wl().will_write();
      init_wl <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      // FP: "9 -> 10;
      while (pipe.in_wl().nitems())
      {
        pipe.out_wl().will_write();
        mark_nodes <<<blocks, threads>>>(gg, prio.gpu_rd_ptr(), pipe.in_wl(), pipe.out_wl());
        pipe.out_wl().will_write();
        drop_marked_nodes_and_nbors <<<blocks, threads>>>(gg, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        STEPS++;
      }
      // FP: "14 -> 15;
    }
  }
  pipe.free();
  // FP: "8 -> 9;
  printf("Total steps: %d\n", STEPS);
  // FP: "15 -> 16;
}