/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
#include "kernels/reduce.cuh"
typedef int edge_data_type;
typedef int node_data_type;
typedef float* gfloat_p;
float* P_CURR ;
float* P_NEXT ;
extern const float ALPHA = 0.85;
extern const float EPSILON = 0.01;
extern int MAX_ITERATIONS ;
static const int __tb_pagerank_main = TB_SIZE;
static const int __tb_remove_dups = TB_SIZE;
__global__ void init_1(CSRGraph graph, float * p_curr, float * residual)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    float update;
    index_type edge_end;
    p_curr[node] = 1.0 - ALPHA;
    update = 1.0/graph.getOutDegree(node);
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      dst = graph.getAbsDestination(edge);
      atomicAdd(residual + dst, update);
    }
  }
  // FP: "10 -> 11;
}
__global__ void init_2(CSRGraph graph, float * residual, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    residual[node] *= (1.0 - ALPHA) * ALPHA;
    (out_wl).push(node);
  }
  // FP: "5 -> 6;
}
__global__ void remove_dups(int * marks, Worklist2 in_wl, Worklist2 out_wl, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  index_type wlnode2_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    marks[node] = wlnode;
  }
  // FP: "6 -> 7;
  gb.Sync();
  // FP: "7 -> 8;
  wlnode2_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode2 = 0 + tid; wlnode2 < wlnode2_end; wlnode2 += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode2, node);
    if (marks[node] == wlnode2)
    {
      (out_wl).push(node);
    }
  }
  // FP: "14 -> 15;
}
__global__ void __launch_bounds__(TB_SIZE, 3) pagerank_main(CSRGraph graph, float * p_curr, float * residual, float * p_diff, Worklist2 in_wl, Worklist2 out_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int sdeg;
    float update;
    int node;
    bool pop;
    float res;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop)
    {
      res =atomicExch(residual + node, 0);
      p_curr[node] += res;
      sdeg = graph.getOutDegree(node);
      update = res * ALPHA / sdeg;
    }
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst;
      float prev;
      dst = graph.getAbsDestination(edge);
      prev = atomicAdd(residual + dst, update);
      if (prev + update > EPSILON && prev < EPSILON)
      {
        (out_wl).push(dst);
      }
    }
  }
  // FP: "17 -> 18;
}
void gg_main(CSRGraph& hg, CSRGraph& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  static GlobalBarrierLifetime remove_dups_barrier;
  static bool remove_dups_barrier_inited;
  PipeContextT<Worklist2> pipe;
  // FP: "1 -> 2;
  Shared<float> p[3] = {Shared<float> (hg.nnodes), Shared<float> (hg.nnodes), Shared<float>(hg.nnodes)};
  // FP: "2 -> 3;
  Shared<float> r (hg.nnodes);
  // FP: "3 -> 4;
  Shared<int> marks (hg.nnodes);
  // FP: "4 -> 5;
  static const size_t remove_dups_residency = maximum_residency(remove_dups, __tb_remove_dups, 0);
  static const size_t remove_dups_blocks = GG_MIN(blocks.x, ggc_get_nSM() * remove_dups_residency);
  if(!remove_dups_barrier_inited) { remove_dups_barrier.Setup(remove_dups_blocks); remove_dups_barrier_inited = true;};
  // FP: "5 -> 6;
  int curr = 0;
  int iter = 0;
  float l1 = 0;
  // FP: "6 -> 7;
  r.zero_gpu();
  // FP: "7 -> 8;
  init_1 <<<blocks, threads>>>(gg, p[0].gpu_wr_ptr(), r.gpu_wr_ptr());
  // FP: "8 -> 9;
  gfloat_p p0 =p[0].gpu_wr_ptr();
  gfloat_p p2 =p[2].gpu_wr_ptr();
  gfloat_p rp =r.gpu_wr_ptr();
  // FP: "9 -> 10;
  pipe = PipeContextT<Worklist2>(hg.nedges);
  {
    {
      pipe.out_wl().will_write();
      init_2 <<<blocks, threads>>>(gg, rp, pipe.in_wl(), pipe.out_wl());
      pipe.in_wl().swap_slots();
      pipe.advance2();
      // FP: "11 -> 12;
      while (pipe.in_wl().nitems())
      {
        pipe.out_wl().will_write();
        pagerank_main <<<blocks, __tb_pagerank_main>>>(gg, p0, rp, p2, pipe.in_wl(), pipe.out_wl());
        pipe.in_wl().swap_slots();
        pipe.advance2();
        iter++;
        if (iter >= MAX_ITERATIONS)
        {
          break;
        }
      }
      // FP: "18 -> 19;
    }
  }
  pipe.free();
  // FP: "10 -> 11;
  printf("PR took %d iterations\n", iter);
  P_CURR = p[0].cpu_rd_ptr();
  P_NEXT = p[0].cpu_rd_ptr();
  // FP: "19 -> 20;
}