/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraphTex &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=texture $ cuda.use_worklist_slots=True $ cuda.worklist_type=texture";
extern int DELTA;
typedef int edge_data_type;
typedef int node_data_type;
typedef int * gint_p;
extern const node_data_type INF = INT_MAX;
static const int __tb_sssp_kernel = TB_SIZE;
static const int __tb_remove_dups = TB_SIZE;
__global__ void kernel(CSRGraphTex graph, int src)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    graph.node_data[node] = (node == src) ? 0 : INF ;
  }
  // FP: "4 -> 5;
}
__global__ void remove_dups(int * marks, WorklistT in_wl, WorklistT out_wl, GlobalBarrier gb)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  index_type wlnode2_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode, node);
    marks[node] = wlnode;
  }
  // FP: "6 -> 7;
  gb.Sync();
  // FP: "7 -> 8;
  wlnode2_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode2 = 0 + tid; wlnode2 < wlnode2_end; wlnode2 += nthreads)
  {
    int node;
    bool pop;
    pop = (in_wl).pop_id(wlnode2, node);
    if (marks[node] == wlnode2)
    {
      (out_wl).push(node);
    }
  }
  // FP: "14 -> 15;
}
__global__ void __launch_bounds__(TB_SIZE, 2) sssp_kernel(CSRGraphTex graph, int delta, WorklistT in_wl, WorklistT out_wl, WorklistT re_wl)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  if (tid == 0)
    in_wl.reset_next_slot();

  index_type wlnode_end;
  // FP: "1 -> 2;
  wlnode_end = *((volatile index_type *) (in_wl).dindex);
  for (index_type wlnode = 0 + tid; wlnode < wlnode_end; wlnode += nthreads)
  {
    int node;
    bool pop;
    index_type edge_end;
    pop = (in_wl).pop_id(wlnode, node);
    if (pop && graph.node_data[node] == INF)
    {
      continue;
    }
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst = graph.getAbsDestination(edge);
      edge_data_type wt = graph.getAbsWeight(edge);
      if (graph.node_data[dst] > graph.node_data[node] + wt)
      {
        atomicMin(graph.node_data + dst, graph.node_data[node] + wt);
        if (graph.node_data[node] + wt <= delta)
        {
          (re_wl).push(dst);
        }
        else
        {
          (out_wl).push(dst);
        }
      }
    }
  }
}
void gg_main(CSRGraphTex& hg, CSRGraphTex& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  static GlobalBarrierLifetime remove_dups_barrier;
  static bool remove_dups_barrier_inited;
  gint_p glevel;
  PipeContextT<WorklistT> pipe;
  // FP: "1 -> 2;
  Shared<int> level (hg.nnodes);
  // FP: "2 -> 3;
  level.cpu_wr_ptr();
  // FP: "3 -> 4;
  static const size_t remove_dups_residency = maximum_residency(remove_dups, __tb_remove_dups, 0);
  static const size_t remove_dups_blocks = GG_MIN(blocks.x, ggc_get_nSM() * remove_dups_residency);
  if(!remove_dups_barrier_inited) { remove_dups_barrier.Setup(remove_dups_blocks); remove_dups_barrier_inited = true;};
  // FP: "4 -> 5;
  // FP: "5 -> 6;
  kernel <<<blocks, threads>>>(gg, 0);
  // FP: "6 -> 7;
  int i = 0;
  int curdelta = 0;
  // FP: "7 -> 8;
  printf("delta: %d\n", DELTA);
  // FP: "8 -> 9;
  glevel = level.gpu_wr_ptr();
  // FP: "9 -> 10;
  pipe = PipeContextT<WorklistT>(gg.nedges*2);
  pipe.in_wl().wl[0] = 0;
  pipe.in_wl().update_gpu(1);
  {
    while (pipe.in_wl().nitems())
    {
      while (pipe.in_wl().nitems())
      {
        pipe.out_wl().will_write();
        pipe.re_wl().will_write();
        sssp_kernel <<<blocks, __tb_sssp_kernel>>>(gg, curdelta, pipe.in_wl(), pipe.out_wl(), pipe.re_wl());
        pipe.in_wl().swap_slots();
        pipe.retry2();
      }
      pipe.advance2();
      pipe.out_wl().will_write();
      remove_dups <<<remove_dups_blocks, __tb_remove_dups>>>(glevel, pipe.in_wl(), pipe.out_wl(), remove_dups_barrier);
      pipe.in_wl().swap_slots();
      pipe.advance2();
      i++;
      curdelta += DELTA;
    }
  }
  pipe.free();
  // FP: "14 -> 15;
  printf("iterations: %d\n", i);
  // FP: "15 -> 16;
}