/*  -*- mode: c++ -*-  */
#include "gg.h"
#include "ggcuda.h"

void kernel_sizing(CSRGraph &, dim3 &, dim3 &);
#define TB_SIZE 256
const char *GGC_OPTIONS = "coop_conv=False $ outline_iterate_gb=False $ backoff_blocking_factor=4 $ parcomb=False $ np_schedulers=set(['fg', 'tb', 'wp']) $ cc_disable=set([]) $ hacks=set([]) $ np_factor=1 $ instrument=set([]) $ unroll=[] $ instrument_mode=None $ read_props=None $ outline_iterate=False $ ignore_nested_errors=False $ np=False $ write_props=None $ quiet_cgen=True $ retry_backoff=False $ cuda.graph_type=basic $ cuda.use_worklist_slots=True $ cuda.worklist_type=basic";
#include <kernels/segmentedsort.cuh>
void debug_output(CSRGraphTy &g, unsigned int *valid_edges);;
__global__ void preprocess(CSRGraph graph, unsigned int * valid_edges)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type node_end;
  // FP: "1 -> 2;
  const index_type last  = graph.nnodes;
  // FP: "2 -> 3;
  node_end = (graph).nnodes;
  for (index_type node = 0 + tid; node < node_end; node += nthreads)
  {
    bool pop;
    int degree;
    index_type edge_end;
    pop = graph.valid_node(node);;
    if (pop)
    {
      degree = graph.getOutDegree(node);
    }
    edge_end = (graph).getFirstEdge((node) + 1);
    for (index_type edge = (graph).getFirstEdge(node) + 0; edge < edge_end; edge += 1)
    {
      index_type dst = graph.getAbsDestination(edge);
      int dst_degree = graph.getOutDegree(dst);
      if ((dst_degree > degree) || (dst_degree == degree && dst > node))
      {
        graph.edge_data[edge] = dst;
        atomicAdd(valid_edges + node, 1);
      }
      else
      {
        graph.edge_data[edge] = graph.nnodes;
      }
    }
  }
  // FP: "16 -> 17;
}
__device__ unsigned int intersect(CSRGraph graph, index_type u, index_type v, unsigned int * valid_edges)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  // FP: "1 -> 2;
  index_type u_start = graph.getFirstEdge(u);
  index_type u_end = u_start + valid_edges[u];
  index_type v_start = graph.getFirstEdge(v);
  index_type v_end = v_start + valid_edges[v];
  int count = 0;
  index_type u_it = u_start;
  index_type v_it = v_start;
  index_type a ;
  index_type b ;
  // FP: "2 -> 3;
  while (u_it < u_end && v_it < v_end)
  {
    a = graph.getAbsDestination(u_it);
    b = graph.getAbsDestination(v_it);
    int d = a - b;
    if (d <= 0)
    {
      u_it++;
    }
    if (d >= 0)
    {
      v_it++;
    }
    if (d == 0)
    {
      count++;
    }
  }
  // FP: "16 -> 17;
  return count;
  // FP: "17 -> 18;
}
__global__ void count_triangles(CSRGraph graph, unsigned int * valid_edges, int * count)
{
  unsigned tid = TID_1D;
  unsigned nthreads = TOTAL_THREADS_1D;

  const unsigned __kernel_tb_size = TB_SIZE;
  index_type v_end;
  // FP: "1 -> 2;
  int lcount =0;
  // FP: "2 -> 3;
  v_end = (graph).nnodes;
  for (index_type v = 0 + tid; v < v_end; v += nthreads)
  {
    bool pop;
    int d_v;
    index_type edge_end;
    pop = graph.valid_node(v);;
    edge_end = (graph).getFirstEdge(v)+ valid_edges[v];
    for (index_type edge = (graph).getFirstEdge(v) + 0; edge < edge_end; edge += 1)
    {
      index_type u = graph.getAbsDestination(edge);
      index_type d_u = graph.getOutDegree(u);
      int xcount = 0;
      xcount = intersect(graph, u, v, valid_edges);
      if (xcount)
      {
        atomicAdd(count, xcount);
      }
    }
  }
  // FP: "13 -> 14;
}
void gg_main(CSRGraphTy& hg, CSRGraphTy& gg)
{
  dim3 blocks, threads;
  kernel_sizing(gg, blocks, threads);
  // FP: "1 -> 2;
  Shared<int> count (1);
  Shared<unsigned int> valid_edges (hg.nnodes);
  // FP: "2 -> 3;
  count.zero_gpu();
  // FP: "3 -> 4;
  valid_edges.zero_gpu();
  // FP: "4 -> 5;
  preprocess <<<blocks, threads>>>(gg, valid_edges.gpu_wr_ptr());
  // FP: "5 -> 6;
  mgpu::SegSortPairsFromIndices(gg.edge_data, gg.edge_dst, gg.nedges, (const int *) gg.row_start + 1, gg.nnodes - 1, *mgc);;
  // FP: "6 -> 7;
  count_triangles <<<blocks, threads>>>(gg, valid_edges.gpu_rd_ptr(), count.gpu_wr_ptr());
  // FP: "7 -> 8;
  printf("triangles: %d\n", *count.cpu_rd_ptr());
  // FP: "8 -> 9;
}